# -*- coding: utf-8 -*-

import re
from collections import OrderedDict, defaultdict

import attr

from . import biblio
from . import config
from . import Line
from .html import *
from .messages import *
from functools import reduce

# This function does a single pass through the doc,
# finding all the "data blocks" and processing them.
# A "data block" is any <pre> or <xmp> element.
#
# When a data block is found, the *contents* of the block
# are passed to the appropriate processing function as an
# array of lines.  The function should return a new array
# of lines to replace the *entire* block.
#
# That is, we give you the insides, but replace the whole
# thing.
#
# Additionally, we pass in the tag-name used (pre or xmp)
# and the line with the content, in case it has useful data in it.


def transformDataBlocks(doc, lines):
    fromStrings = False
    if any(isinstance(l, str) for l in lines):
        fromStrings = True
        lines = [Line.Line(-1, l) for l in lines]
    inBlock = False
    blockTypes = {
        'propdef': transformPropdef,
        'descdef': transformDescdef,
        'elementdef': transformElementdef,
        'argumentdef': transformArgumentdef,
        'railroad': transformRailroad,
        'biblio': transformBiblio,
        'anchors': transformAnchors,
        'link-defaults': transformLinkDefaults,
        'ignored-specs': transformIgnoredSpecs,
        'info': transformInfo,
        'include': transformInclude,
        'include-code': transformIncludeCode,
        'include-raw': transformIncludeRaw,
        'pre': transformPre
    }
    blockType = ""
    tagName = ""
    blockLines = []
    newLines = []
    for line in lines:
        # Look for the start of a block.
        match = re.match(r"\s*<(pre|xmp)[\s>]", line.text, re.I)
        # Note that, by design, I don't pay attention to anything on the same line as the start tag,
        # unless it's single-line.
        if match and not inBlock:
            inBlock = True
            tagName = match.group(1)
            blockClasses = classesFromLine(line)
            seenClasses = []
            for t in blockTypes.keys():
                if t in blockClasses:
                    seenClasses.append(t)
            if not seenClasses:
                blockType = "pre"
            elif len(seenClasses) == 1:
                blockType = seenClasses[0]
            else:
                die("Found {0} classes on the <{1}>, so can't tell which to process the block as. Please use only one.", config.englishFromList(("'{0}'".format(x) for x in seenClasses), "and"), tagName, lineNum=line.i)
                blockType = "pre"
        # Look for the end of a block.
        match = re.match(r"(.*)</" + tagName + ">(.*)", line.text, re.I)
        if match and inBlock:
            inBlock = False
            if len(blockLines) == 0:
                # Single-line <pre>.
                match = re.match(r"(\s*<{0}[^>]*>)(.*)</{0}>(.*)".format(tagName), line.text, re.I)
                if not match:
                    die("Can't figure out how to parse this datablock line:\n{0}", line.text, lineNum=line.i)
                    blockLines = []
                    continue
                repl = blockTypes[blockType](
                    lines=[match.group(2)],
                    tagName=tagName,
                    firstLine=match.group(1),
                    lineNum=line.i,
                    doc=doc)
                newLines.extend(Line.Line(line.i,l) for l in repl)
                line.text = match.group(3)
                newLines.append(line)
            elif re.match(r"^\s*$", match.group(1)):
                # End tag was the first tag on the line.
                # Remove the tag from the line.
                repl = blockTypes[blockType](
                    lines=cleanPrefix([l.text for l in blockLines[1:]], doc.md.indent),
                    tagName=tagName,
                    firstLine=blockLines[0].text,
                    lineNum=blockLines[0].i,
                    doc=doc)
                newLines.extend(Line.Line(blockLines[0].i, l) for l in repl)
                line.text = match.group(2)
                newLines.append(line)
            else:
                # End tag was at the end of line of useful content.
                # Process the stuff before it, preserve the stuff after it.
                repl = blockTypes[blockType](
                    lines=cleanPrefix([l.text for l in blockLines[1:]] + [match.group(1)], doc.md.indent),
                    tagName=tagName,
                    firstLine=blockLines[0].text,
                    lineNum=blockLines[0].i,
                    doc=doc)
                newLines.extend(Line.Line(blockLines[0].i, l) for l in repl)
                line.text = match.group(2)
                newLines.append(line)
            tagName = ""
            blockType = ""
            blockLines = []
            continue
        elif inBlock:
            blockLines.append(line)
        else:
            newLines.append(line)

    #for line in newLines:
    #    print line

    if fromStrings:
        return [l.text for l in newLines]
    else:
        return newLines

def cleanPrefix(lines, tabSize):
    # Remove the longest common whitespace prefix from the lines.
    # Returns a fresh array, does not mutate the passed lines.
    if not lines:
        return []
    prefix = reduce(commonPrefix, map(getWsPrefix, lines))
    prefixLen = len(prefix)
    return [line[prefixLen:] for line in lines]

def commonPrefix(line1, line2):
    prefixSoFar = ""
    for i,char in enumerate(line1):
        if i == len(line2):
            break
        if char == line2[i]:
            prefixSoFar += char
        else:
            break
    return prefixSoFar

def getWsPrefix(line):
    return re.match(r"(\s*)", line).group(1)


'''
When writing a new transformFoo function,
PAY ATTENTION TO THE INDENT OF THE FIRST LINE
OR ELSE YOU'LL MESS UP THE MARKDOWN PARSER.
Generally, just spam the first-line's indent onto every line;
if you're outputting a raw element (<pre>/etc),
just spam it onto the first line.
'''

def transformPre(lines, tagName, firstLine, **kwargs):
    # If the last line in the source is a </code></pre>,
    # the generic processor will turn that into a final </code> line,
    # which'll mess up the indent finding.
    # Instead, specially handle this case.
    if len(lines) == 0:
        return [firstLine, "</{0}>".format(tagName)]

    if re.match(r"\s*</code>\s*$", lines[-1]):
        lastLine = "</code></{0}>".format(tagName)
        lines = lines[:-1]
    else:
        lastLine = "</{0}>".format(tagName)

    if len(lines) == 0:
        return [firstLine, lastLine]

    indent = float("inf")
    for (i, line) in enumerate(lines):
        if line.strip() == "":
            continue

        # Use tabs in the source, but spaces in the output,
        # because tabs are ginormous in HTML.
        lines[i] = lines[i].replace("\t", "  ")

        # Find the line with the shortest whitespace prefix.
        # (It might not be the first!)
        indent = min(indent, len(re.match(r" *", lines[i]).group(0)))

    if indent == float("inf"):
        indent = 0

    # Strip off the whitespace prefix from each line
    for (i, line) in enumerate(lines):
        if line.strip() == "":
            continue
        lines[i] = lines[i][indent:]
    # Put the first/last lines back into the results.
    lines[0] = firstLine.rstrip() + lines[0]
    lines.append(lastLine)
    return lines


def transformPropdef(lines, doc, firstLine, lineNum=None, **kwargs):
    attrs = OrderedDict()
    parsedAttrs = parseDefBlock(lines, "propdef")
    # Displays entries in the order specified in attrs,
    # then if there are any unknown parsedAttrs values,
    # they're displayed afterward in the order they were specified.
    # attrs with a value of None are required to be present in parsedAttrs;
    # attrs with any other value are optional, and use the specified value if not present in parsedAttrs
    forHint = ""
    if "Name" in parsedAttrs:
        forHint = " data-link-for-hint='{0}'".format(parsedAttrs["Name"].split(",")[0].strip())
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    if "partial" in firstLine or "New values" in parsedAttrs:
        attrs["Name"] = None
        attrs["New values"] = None
        ret = ["<table class='def propdef partial'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)]
    elif "shorthand" in firstLine:
        attrs["Name"] = None
        attrs["Value"] = None
        for defaultKey in ["Initial", "Applies to", "Inherited", "Percentages", "Computed value", "Animation type"]:
            attrs[defaultKey] = "see individual properties"
        attrs["Canonical order"] = "per grammar"
        ret = ["<table class='def propdef'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)]
    else:
        attrs["Name"] = None
        attrs["Value"] = None
        attrs["Initial"] = None
        attrs["Applies to"] = "all elements"
        attrs["Inherited"] = None
        attrs["Percentages"] = "n/a"
        attrs["Computed value"] = "as specified"
        attrs["Canonical order"] = "per grammar"
        attrs["Animation type"] = "discrete"
        ret = ["<table class='def propdef'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)]
    # We are in the process of migrating specs from using 'Animatable' to
    # using 'Animation type'. If we find 'Animatable' in the parsed attributes,
    # drop the default 'Animation type' entry.
    if "Animatable" in parsedAttrs:
        attrs.pop("Animation type")

    attrsToPrint = []
    for key, val in attrs.items():
        if key in parsedAttrs:
            # Key was provided
            val = parsedAttrs[key]
        elif val is None:
            # Required key, not provided
            die("The propdef for '{0}' is missing a '{1}' line.", parsedAttrs.get("Name", "???"), key, lineNum=lineNum)
            continue
        else:
            # Optional key, just use default
            pass
        attrsToPrint.append((key,val))
    for key,val in parsedAttrs.items():
        # Find any "custom" provided keys
        if key not in attrs:
            attrsToPrint.append((key,val))

    for key,val in attrsToPrint:
        tr = "<tr>"
        th = "<th>{0}:".format(key)
        td = "<td>{0}".format(val)
        if key in ("Value", "New values"):
            tr = "<tr class=value>"
            th = "<th><a href='https://www.w3.org/TR/css-values/#value-defs'>{0}:</a>".format(key)
            td = "<td class=prod>{0}".format(val)
        elif key == "Initial":
            th = "<th><a href='https://www.w3.org/TR/css-cascade/#initial-values'>{0}:</a>".format(key)
        elif key == "Inherited":
            th = "<th><a href='https://www.w3.org/TR/css-cascade/#inherited-property'>{0}:</a>".format(key)
        elif key == "Percentages":
            th = "<th><a href='https://www.w3.org/TR/css-values/#percentages'>{0}:</a>".format(key)
        elif key == "Computed value":
            th = "<th><a href='https://www.w3.org/TR/css-cascade/#computed'>{0}:</a>".format(key)
        elif key in ("Animatable", "Animation type"):
            th = "<th><a href='https://www.w3.org/TR/web-animations/#animation-type'>{0}:</a>".format(key)
        elif key == "Applies to" and val.lower() == "all elements":
            td = "<td><a href='https://www.w3.org/TR/css-pseudo/#generated-content' title='Includes ::before and ::after pseudo-elements.'>all elements</a>"
        ret.append(tr+th+td)
    ret.append("</table>")

    indent = getWsPrefix(firstLine)
    ret = [indent+l for l in ret]

    return ret

# TODO: Make these functions match transformPropdef's new structure


def transformDescdef(lines, doc, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    vals = parseDefBlock(lines, "descdef")
    if "partial" in firstLine or "New values" in vals:
        requiredKeys = ["Name", "For"]
        ret = ["<table class='def descdef partial' data-dfn-for='{0}'{lineNumAttr}>".format(vals.get("For", ""), lineNumAttr=lineNumAttr)]
    if "mq" in firstLine:
        requiredKeys = ["Name", "For", "Value"]
        ret = ["<table class='def descdef mq' data-dfn-for='{0}'{lineNumAttr}>".format(vals.get("For",""), lineNumAttr=lineNumAttr)]
    else:
        requiredKeys = ["Name", "For", "Value", "Initial"]
        ret = ["<table class='def descdef' data-dfn-for='{0}'{lineNumAttr}>".format(vals.get("For", ""), lineNumAttr=lineNumAttr)]
    for key in requiredKeys:
        if key == "For":
            ret.append("<tr><th>{0}:<td><a at-rule>{1}</a>".format(key, vals.get(key,'')))
        elif key == "Value":
            ret.append("<tr><th>{0}:<td class='prod'>{1}".format(key, vals.get(key,'')))
        elif key in vals:
            ret.append("<tr><th>{0}:<td>{1}".format(key, vals.get(key,'')))
        else:
            die("The descdef for '{0}' is missing a '{1}' line.", vals.get("Name", "???"), key, lineNum=lineNum)
            continue
    for key, val in vals.items():
        if key in requiredKeys:
            continue
        ret.append("<tr><th>{0}:<td>{1}".format(key, vals[key]))
    ret.append("</table>")

    indent = getWsPrefix(firstLine)
    ret = [indent+l for l in ret]

    return ret


def transformElementdef(lines, doc, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    attrs = OrderedDict()
    parsedAttrs = parseDefBlock(lines, "elementdef")
    if "Attribute groups" in parsedAttrs or "Attributes" in parsedAttrs:
        html = "<ul>"
        if "Attribute groups" in parsedAttrs:
            groups = [x.strip() for x in parsedAttrs["Attribute groups"].split(",")]
            for group in groups:
                html += "<li><a dfn data-element-attr-group{lineNumAttr}>{0}</a>".format(group, lineNumAttr=lineNumAttr)
            del parsedAttrs["Attribute groups"]
        if "Attributes" in parsedAttrs:
            atts = [x.strip() for x in parsedAttrs["Attributes"].split(",")]
            for att in atts:
                html += "<li><a element-attr for='{1}'{lineNumAttr}>{0}</a>".format(att, parsedAttrs.get("Name", ""), lineNumAttr=lineNumAttr)
        html += "</ul>"
        parsedAttrs["Attributes"] = html

    # Displays entries in the order specified in attrs,
    # then if there are any unknown parsedAttrs values,
    # they're displayed afterward in the order they were specified.
    # attrs with a value of None are required to be present in parsedAttrs;
    # attrs with any other value are optional, and use the specified value if not present in parsedAttrs
    attrs["Name"] = None
    attrs["Categories"] = None
    attrs["Contexts"] = None
    attrs["Content model"] = None
    attrs["Attributes"] = None
    attrs["Dom interfaces"] = None
    ret = ["<table class='def elementdef'{lineNumAttr}>".format(lineNumAttr=lineNumAttr)]
    for key, val in attrs.items():
        if key in parsedAttrs or val is not None:
            if key in parsedAttrs:
                val = parsedAttrs[key]
            if key == "Name":
                ret.append("<tr><th>Name:<td>")
                ret.append(', '.join("<dfn element{lineNumAttr}>{0}</dfn>".format(x.strip(), lineNumAttr=lineNumAttr) for x in val.split(",")))
            elif key == "Content model":
                ret.append("<tr><th>{0}:<td>".format(key))
                ret.extend(val.split("\n"))
            elif key == "Categories":
                ret.append("<tr><th>Categories:<td>")
                ret.append(', '.join("<a dfn{lineNumAttr}>{0}</a>".format(x.strip(), lineNumAttr=lineNumAttr) for x in val.split(",")))
            elif key == "Dom interfaces":
                ret.append("<tr><th>DOM Interfaces:<td>")
                ret.append(', '.join("<a interface{lineNumAttr}>{0}</a>".format(x.strip(), lineNumAttr=lineNumAttr) for x in val.split(",")))
            else:
                ret.append("<tr><th>{0}:<td>{1}".format(key, val))
        else:
            die("The elementdef for '{0}' is missing a '{1}' line.", parsedAttrs.get("Name", "???"), key, lineNum=lineNum)
            continue
    for key, val in parsedAttrs.items():
        if key in attrs:
            continue
        ret.append("<tr><th>{0}:<td>{1}".format(key, val))
    ret.append("</table>")

    indent = getWsPrefix(firstLine)
    ret = [indent+l for l in ret]

    return ret


def transformArgumentdef(lines, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    attrs = parseDefBlock(lines, "argumentdef", capitalizeKeys=False, lineNum=lineNum)
    el = parseHTML(firstLine + "</pre>")[0]
    if "for" in el.attrib:
        forValue = el.get('for')
        el.set("data-dfn-for", forValue)
        if "/" in forValue:
            interface, method = forValue.split("/")
        else:
            die("Argumentdef for='' values need to specify interface/method(). Got '{0}'.", forValue, lineNum=lineNum)
            return []
        removeAttr(el, "for")
    else:
        die("Argumentdef blocks need a for='' attribute specifying their method.", lineNum=lineNum)
        return []
    addClass(el, "data")
    rootAttrs = " ".join("{0}='{1}'".format(k,escapeAttr(v)) for k,v in el.attrib.items())
    text = ('''
<table {attrs}{lineNumAttr}>
    <caption>Arguments for the <a idl lt='{method}' for='{interface}'{lineNumAttr}>{interface}.{method}</a> method.</caption>
    <thead>
        <tr>
            <th>Parameter
            <th>Type
            <th>Nullable
            <th>Optional
            <th>Description
    <tbody>'''.format(attrs=rootAttrs, interface=interface, method=method, lineNumAttr=lineNumAttr)
+ '\n'.join(['''
        <tr>
            <td><dfn argument{lineNumAttr}>{0}</dfn>
            <td>
            <td>
            <td>
            <td>{1}'''.format(param, desc, lineNumAttr=lineNumAttr)
    for param,desc in attrs.items()])
+ '''
</table>''')

    indent = getWsPrefix(firstLine)
    lines = [indent+line for line in text.split("\n")]

    return lines


def parseDefBlock(lines, type, capitalizeKeys=True, lineNum=None):
    vals = OrderedDict()
    lastKey = None
    for line in lines:
        match = re.match(r"\s*([^:]+):\s*(\S.*)?", line)
        if match is None:
            if lastKey is not None and (line.strip() == "" or re.match(r"\s+", line)):
                key = lastKey
                val = line.strip()
            else:
                die("Incorrectly formatted {2} line for '{0}':\n{1}", vals.get("Name", "???"), line, type, lineNum=lineNum)
                continue
        else:
            key = match.group(1).strip()
            if capitalizeKeys:
                key = key.capitalize()
            lastKey = key
            val = (match.group(2) or "").strip()
        if key in vals:
            vals[key] += "\n" + val
        else:
            vals[key] = val
    return vals


def transformRailroad(lines, doc, firstLine, **kwargs):
    import io
    from . import railroadparser
    ret = ["<div class='railroad'>"]
    doc.extraStyles['style-railroad'] = "svg.railroad-diagram{background-color:hsl(30,20%,95%);}svg.railroad-diagram path{stroke-width:3px;stroke:black;fill:rgba(0,0,0,0);}svg.railroad-diagram text{font:bold 14px monospace;text-anchor:middle;}svg.railroad-diagram text.label{text-anchor:start;}svg.railroad-diagram text.comment{font:italic 12px monospace;}svg.railroad-diagram rect{stroke-width:3px;stroke:black;fill:hsl(120,100%,90%);}"
    code = ''.join(lines)
    diagram = railroadparser.parse(code)
    if diagram:
        temp = io.StringIO()
        diagram.writeSvg(temp.write)
        ret.append(temp.getvalue())
        temp.close()
        ret.append("</div>")

        indent = getWsPrefix(firstLine)
        ret = [indent+l for l in ret]

        return ret
    else:
        return []


def transformBiblio(lines, doc, **kwargs):
    storage = defaultdict(list)
    biblio.processSpecrefBiblioFile(''.join(lines), storage, order=1)
    for k,vs in storage.items():
        doc.refs.biblioKeys.add(k)
        doc.refs.biblios[k].extend(vs)
    return []


def transformAnchors(lines, doc, lineNum=None, **kwargs):
    anchors = parseInfoTree(lines, doc.md.indent, lineNum)
    processAnchors(anchors, doc, lineNum)
    return []


def processAnchors(anchors, doc, lineNum=None):
    for anchor in anchors:
        if "type" not in anchor or len(anchor['type']) != 1:
            die("Each anchor needs exactly one type. Got:\n{0}", config.printjson(anchor), lineNum=lineNum)
            continue
        if "text" not in anchor or len(anchor['text']) != 1:
            die("Each anchor needs exactly one text. Got:\n{0}", config.printjson(anchor), lineNum=lineNum)
            continue
        if "url" not in anchor and "urlPrefix" not in anchor:
            die("Each anchor needs a url and/or at least one urlPrefix. Got:\n{0}", config.printjson(anchor), lineNum=lineNum)
            continue
        if "urlPrefix" in anchor:
            urlPrefix = ''.join(anchor['urlPrefix'])
        else:
            urlPrefix = ""
        if "url" in anchor:
            urlSuffix = anchor['url'][0]
        else:
            urlSuffix = config.simplifyText(anchor['text'][0])
        url = urlPrefix + ("" if "#" in urlPrefix or "#" in urlSuffix else "#") + urlSuffix
        status = "local"
        shortname = None
        level = None
        if "shortname" in anchor and "level" in anchor:
            shortname = anchor['shortname'][0]
            level = anchor['level'][0]
        spec = anchor["spec"][0] if "spec" in anchor else None
        if shortname and not spec:
            if level:
                spec = "{0}-{1}".format(shortname, level)
            else:
                spec = shortname
        elif spec and not shortname:
            match = re.match("(.*)-(\d+)$", spec)
            if match:
                shortname = match.group(1)
                level = match.group(2)
            else:
                shortname = spec
                level = ""
        if "status" in anchor:
            status = anchor["status"][0]
            if status in config.linkStatuses:
                pass
            else:
                die("Anchor statuses must be {1}. Got '{0}'.", status, config.englishFromList(config.linkStatuses), lineNum=lineNum)
                continue
        else:
            status = "anchor-block"
        if anchor['type'][0] in config.lowercaseTypes:
            anchor['text'][0] = anchor['text'][0].lower()
        doc.refs.anchorBlockRefs._refs[anchor['text'][0]].append({
            "linkingText": anchor['text'][0],
            "type": anchor['type'][0].lower(),
            "url": url,
            "shortname": shortname.lower() if shortname is not None else doc.md.shortname,
            "level": level if level is not None else doc.md.level,
            "for": anchor.get('for', []),
            "export": True,
            "status": status,
            "spec": spec.lower() if spec is not None else ''
        })
        methodishStart = re.match(r"([^(]+\()[^)]", anchor['text'][0])
        if methodishStart:
            doc.refs.anchorBlockRefs.addMethodVariants(anchor['text'][0], anchor.get('for', []), doc.md.shortname)


def transformLinkDefaults(lines, doc, lineNum=None, **kwargs):
    lds = parseInfoTree(lines, doc.md.indent, lineNum)
    processLinkDefaults(lds, doc, lineNum)
    return []


def processLinkDefaults(lds, doc, lineNum=None):
    for ld in lds:
        if len(ld.get('type', [])) != 1:
            die("Every link default needs exactly one type. Got:\n{0}", config.printjson(ld), lineNum=lineNum)
            continue
        else:
            type = ld['type'][0]
        if len(ld.get('spec', [])) != 1:
            die("Every link default needs exactly one spec. Got:\n{0}", config.printjson(ld), lineNum=lineNum)
            continue
        else:
            spec = ld['spec'][0]
        if len(ld.get('text', [])) != 1:
            die("Every link default needs exactly one text. Got:\n{0}", config.printjson(ld), lineNum=lineNum)
            continue
        else:
            text = ld['text'][0]
        if 'for' in ld:
            for _for in ld['for']:
                doc.md.linkDefaults[text].append((spec, type, ld.get('status', None), _for))
        else:
            doc.md.linkDefaults[text].append((spec, type, ld.get('status', None), None))


def transformIgnoredSpecs(lines, doc, lineNum=None, **kwargs):
    specs = parseInfoTree(lines, doc.md.indent, lineNum)
    processIgnoredSpecs(specs, doc, lineNum)
    return []


def processIgnoredSpecs(specs, doc, lineNum=None):
    for spec in specs:
        if len(spec.get('spec', [])) == 0:
            die("Every ignored spec line needs at least one 'spec' value. Got:\n{0}", config.printjson(spec), lineNum=lineNum)
            continue
        specNames = spec.get('spec')
        if len(spec.get('replacedBy', [])) > 1:
            die("Every ignored spec line needs at most one 'replacedBy' value. Got:\n{0}", config.printjson(spec), lineNum=lineNum)
            continue
        replacedBy = spec.get('replacedBy')[0] if 'replacedBy' in spec else None
        for specName in specNames:
            if replacedBy:
                doc.refs.replacedSpecs.add((specName, replacedBy))
            else:
                doc.refs.ignoredSpecs.add(specName)


def transformInfo(lines, doc, lineNum=None, **kwargs):
    # More generic InfoTree system.
    # A <pre class=info> can contain any of the InfoTree collections,
    # identified by an 'info' line.
    infos = parseInfoTree(lines, doc.md.indent, lineNum)
    processInfo(infos, doc, lineNum)
    return []


def processInfo(infos, doc, lineNum=None):
    knownInfoTypes = {
        "anchors": processAnchors,
        "link-defaults": processLinkDefaults,
        "ignored-specs": processIgnoredSpecs
    }
    infoCollections = defaultdict(list)
    for info in infos:
        if len(info.get('info', [])) != 1:
            die("Every info-block line needs exactly one 'info' type. Got:\n{0}", config.printjson(info), lineNum=lineNum)
            continue
        infoType = info.get('info')[0].lower()
        if infoType not in knownInfoTypes:
            die("Unknown info-block type '{0}'", infoType, lineNum=lineNum)
            continue
        infoCollections[infoType].append(info)
    for infoType, infos in infoCollections.items():
        knownInfoTypes[infoType](infos, doc, lineNum=0)


def transformInclude(lines, doc, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    infos = parseInfoTree(lines, doc.md.indent, lineNum)
    path = None
    macros = {}
    for info in infos:
        if "path" in info:
            if path is None:
                path = info['path'][0]
            else:
                die("Include blocks must only contain a single 'path'.", lineNum=lineNum)
        if "macros" in info:
            for k,v in info.items():
                if k == "macros":
                    continue
                if k not in macros and len(v) == 1:
                    macros[k] = v[0]
                else:
                    die("Include block defines the '{0}' local macro more than once.", k, lineNum=lineNum)
    if path:
        el = "<pre class=include path='{0}'".format(escapeAttr(path))
        for i,(k,v) in enumerate(macros.items()):
            el += " macro-{0}='{1} {2}'".format(i, k, escapeAttr(v))
        el += "{lineNumAttr}></pre>".format(lineNumAttr=lineNumAttr)

        indent = getWsPrefix(firstLine)
        return [indent+el]
    else:
        return []


def transformIncludeCode(lines, doc, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    infos = parseInfoTree(lines, doc.md.indent, lineNum)
    path = None
    highlight = None
    lineStart = None
    show = []
    lineHighlight = []
    lineNumbers = False
    for info in infos:
        if "path" in info:
            if path is None:
                path = info['path'][0]
            else:
                die("Include-code blocks must only contain a single 'path'.", lineNum=lineNum)
        if "highlight" in info:
            if highlight is None:
                highlight = info['highlight'][0]
            else:
                die("Include-code blocks must only contain a single 'highlight'.", lineNum=lineNum)
        if "line-start" in info:
            if lineStart is None:
                lineStart = info['line-start'][0]
            else:
                die("Include-code blocks must only contain a single 'line-start'.", lineNum=lineNum)
        if "show" in info:
            show.extend(info['show'])
        if "line-highlight" in info:
            lineHighlight.extend(info['line-highlight'])
        if "line-numbers" in info:
            lineNumbers = True
        if "no-line-numbers" in info:
            lineNumbers = False


    if path:
        attrs = lineNumAttr
        attrs += " path='{0}'".format(escapeAttr(path))
        if highlight:
            attrs += " highlight='{0}'".format(escapeAttr(highlight))
        if lineStart:
            attrs += " line-start='{0}'".format(escapeAttr(lineStart))
        if show:
            attrs += " data-code-show='{0}'".format(escapeAttr(",".join(show)))
        if lineHighlight:
            attrs += " line-highlight='{0}'".format(escapeAttr(",".join(lineHighlight)))
        if lineNumbers:
            attrs += " line-numbers"
        el = "<pre class=include-code{0}></pre>".format(attrs)
        indent = getWsPrefix(firstLine)
        return [indent+el]
    else:
        return []


def transformIncludeRaw(lines, doc, firstLine, lineNum=None, **kwargs):
    lineNumAttr = ""
    if lineNum is not None:
        lineNumAttr = " line-number={0}".format(lineNum)
    infos = parseInfoTree(lines, doc.md.indent, lineNum)
    path = None
    highlight = None
    lineStart = None
    show = []
    lineHighlight = []
    lineNumbers = False
    for info in infos:
        if "path" in info:
            if path is None:
                path = info['path'][0]
            else:
                die("Include-raw blocks must only contain a single 'path'.", lineNum=lineNum)

    if path:
        attrs = lineNumAttr
        attrs += " path='{0}'".format(escapeAttr(path))
        el = "<pre class=include-raw{0}></pre>".format(attrs)
        indent = getWsPrefix(firstLine)
        return [indent+el]
    else:
        return []



def parseInfoTree(lines, indent=4, lineNum=0):
    # Parses sets of info, which can be arranged into trees.
    # Each info is a set of key/value pairs, semicolon-separated:
    # key1: val1; key2: val2; key3: val3
    # Intead of semicolon-separating, pieces can be nested with higher indentation
    # key1: val1
    #     key2: val2
    #         key3: val3
    # Multiple fragments can be chained off of a single higher-level piece,
    # to avoid repetition:
    # key1: val1
    #     key2: val2
    #     key2a: val2a
    # ===
    # key1: val1; key2: val2
    # key1: val1; key2a: val2a
    # Starting a line with # will comment it out.

    def extendData(datas, infoLevels):
        if not infoLevels:
            return
        newData = defaultdict(list)
        for infos in infoLevels:
            for k,v in infos.items():
                newData[k].extend(v)
        datas.append(newData)

    # Determine the indents, separate the lines.
    datas = []
    infoLevels = []
    lastIndent = -1
    indentSpace = " " * indent
    for i,line in enumerate(lines):
        if lineNum is not None:
            thisLine = int(lineNum) + i + 1
        else:
            thisLine = None
        if line.strip() == "":
            continue
        ws, text = re.match("(\s*)(.*)", line).groups()
        if text.startswith("#"): # comment
            continue
        wsLen = len(ws.replace("\t", indentSpace))
        if wsLen % indent != 0:
            visibleWs = ws.replace("\t", "\\t").replace(" ", "\\s")
            die("Line has inconsistent indentation; use tabs or {1} spaces:\n{0}", visibleWs+text, indent, lineNum=thisLine)
            return []
        wsLen = wsLen // indent
        if wsLen >= lastIndent + 2:
            die("Line jumps {1} indent levels:\n{0}", text, wsLen - lastIndent, lineNum=thisLine)
            return []
        if wsLen <= lastIndent:
            # Previous line was a leaf node; build its full data and add to the list
            extendData(datas, infoLevels[:lastIndent + 1])
        # Otherwise, chained data. Parse it, put it into infoLevels
        info = defaultdict(list)
        for piece in text.split(";"):
            if piece.strip() == "":
                continue
            match = re.match("([^:]+):\s*(.*)", piece)
            if not match:
                die("Line doesn't match the grammar `k:v; k:v; k:v`:\n{0}", line, lineNum=thisLine)
                return []
            key = match.group(1).strip()
            val = match.group(2).strip()
            info[key].append(val)
        if wsLen < len(infoLevels):
            infoLevels[wsLen] = info
        else:
            infoLevels.append(info)
        lastIndent = wsLen
    # Grab the last bit of data.
    extendData(datas, infoLevels[:lastIndent + 1])
    return datas


def classesFromLine(line):
    tag = parseTag(line.text.strip(), lineNumber=line.i)
    if tag is None:
        return set()
    if "class" not in tag.attrs:
        return set()
    return set(tag.attrs["class"].strip().split())

@attr.s(slots=True)
class StartTag(object):
    tag = attr.ib()
    attrs = attr.ib(default=attr.Factory(dict))

def parseTag(text, lineNumber):
    '''
    Parses a tag from a string,
    conformant to the HTML parsing algorithm.
    The text must start with the opening < character.
    '''

    def parseerror(index, state):
        die("Tried to parse a start tag from '{0}', but failed at character {1} '{2}' and parse-state '{3}'.", text, index, text[index], state, lineNum=lineNumber)
        return
    def eof(i,text):
        return i >= len(text)
    i = 0
    state = "data"
    while True:
        if eof(i, text):
            parseerror(i, state)
            return
        if state == "data":
            if text[i] == "<":
                state = "tag-open"
                i += 1
                continue
            else:
                parseerror(i, state)
                return
        elif state == "tag-open":
            if text[i].isalpha():
                state = "tag-name"
                continue
            else:
                parseerror(i, state)
                return
        elif state == "tag-name":
            tagname = ""
            while not eof(i,text) and re.match(r"[^\s/>]", text[i]):
                tagname += text[i].lower()
                i += 1
            tag = StartTag(tagname)
            if text[i] == ">":
                return tag
            elif text[i] == "/":
                state = "self-closing-start-tag"
                i += 1
                continue
            elif text[i].isspace():
                state = "before-attribute-name"
                i += 1
                continue
            else:
                parseerror(i, state)
                return
        elif state == "self-closing-start-tag":
            if text[i] == ">":
                return tag
            else:
                parseerror(i, state)
                return
        elif state == "before-attribute-name":
            if text[i].isspace():
                i += 1
                continue
            elif text[i] == "/" or text[i] == ">":
                state = "after-attribute-name"
                continue
            elif text[i] == "=":
                parseerror(i, state)
                return
            else:
                state = "attribute-name"
                continue
        elif state == "attribute-name":
            attrName = ""
            while not eof(i,text) and re.match(r"[^\s/>=\"'<]", text[i]):
                attrName += text[i]
                i += 1
            tag.attrs[attrName] = ""
            if text[i].isspace() or text[i] == "/" or text[i] == ">":
                state = "after-attribute-name"
                continue
            elif text[i] == "=":
                state = "before-attribute-value"
                i += 1
                continue
            else:
                parseerror(i,state)
                return
        elif state == "after-attribute-name":
            if text[i].isspace():
                i += 1
                continue
            elif text[i] == "/":
                state = "self-closing-start-tag"
                i += 1
                continue
            elif text[i] == "=":
                state = "before-attribute-value"
                i += 1
                continue
            elif text[i] == ">":
                return tag
            else:
                state = "attribute-name"
                continue
        elif state == "before-attribute-value":
            if text[i].isspace():
                i += 1
                continue
            elif text[i] == '"':
                state = "attribute-value-double-quoted"
                i += 1
                continue
            elif text[i] == "'":
                state = "attribute-value-single-quoted"
                i += 1
                continue
            elif text[i] == "=":
                parseerror(i,state)
                return
            else:
                state = "attribute-value-unquoted"
                continue
        elif state == "attribute-value-double-quoted":
            attrValue = ""
            while not eof(i,text) and not text[i] == '"':
                attrValue += text[i]
                i += 1
            tag.attrs[attrName] = attrValue
            if text[i] == '"':
                state = "after-attribute-value-quoted"
                i += 1
                continue
            else:
                parseerror(i,state)
                return
        elif state == "attribute-value-single-quoted":
            attrValue = ""
            while not eof(i,text) and not text[i] == "'":
                attrValue += text[i]
                i += 1
            tag.attrs[attrName] = attrValue
            if text[i] == "'":
                state = "after-attribute-value-quoted"
                i += 1
                continue
            else:
                parseerror(i,state)
                return
        elif state == "attribute-value-unquoted":
            attrValue = ""
            while not eof(i,text) and re.match("[^\s<>'\"=`]", text[i]):
                attrValue += text[i]
                i += 1
            tag.attrs[attrName] = attrValue
            if text[i].isspace():
                state = "before-attribute-name"
                i += 1
                continue
            elif text[i] == ">":
                return tag
            else:
                parseerror(i,state)
                return
        elif state == "after-attribute-value-quoted":
            if text[i].isspace():
                state = "before-attribute-name"
                i += 1
                continue
            elif text[i] == "/":
                state = "self-closing-start-tag"
                i += 1
                continue
            elif text[i] == ">":
                return tag
            else:
                parseerror(i,state)
                return



# -*- coding: utf-8 -*-

import collections
import itertools
import re
from . import config
from .html import *
from .messages import *


def loadCSSLexer():
    from .lexers import CSSLexer
    return CSSLexer()
customLexers = {
    "css": loadCSSLexer
}

ColoredText = collections.namedtuple('ColoredText', ['text', 'color'])

def addSyntaxHighlighting(doc):
    if doc.md.slimBuildArtifact:
        return
    normalizeHighlightMarkers(doc)

    # Highlight all the appropriate elements
    highlightingOccurred = False
    lineWrappingOccurred = False
    lineHighlightingOccurred = False
    for el in findAll("xmp, pre, code", doc):
        # Find whether to highlight, and what the lang is
        lang = determineHighlightLang(doc, el)
        if lang is False:
            # Element was already highlighted, but needs styles
            highlightingOccurred = True
        elif lang:
            highlightEl(el, lang)
            highlightingOccurred = True
        # Find whether to add line numbers
        addLineNumbers, lineStart, lineHighlights = determineLineNumbers(doc, el)
        if addLineNumbers or lineHighlights:
            addLineWrappers(el, numbers=addLineNumbers, start=lineStart, highlights=lineHighlights)
            if addLineNumbers:
                lineWrappingOccurred = True
            if lineHighlights:
                lineHighlightingOccurred = True

    if highlightingOccurred:
        doc.extraStyles['style-syntax-highlighting'] += getHighlightStyles()
    if lineWrappingOccurred:
        doc.extraStyles['style-line-numbers'] += getLineNumberStyles()
    if lineHighlightingOccurred:
        doc.extraStyles['style-line-highlighting'] += getLineHighlightingStyles()


def determineHighlightLang(doc, el):
    # Either returns a normalized highlight lang,
    # False indicating the element was already highlighted,
    # or None indicating the element shouldn't be highlighted.
    attr, lang = closestAttr(el, "nohighlight", "highlight")
    lang = normalizeLanguageName(lang)
    if lang == "webidl" and el.tag == "code" and parentElement(el).tag == "dfn":
        # No such thing as a dfn that needs to be WebIDL-highlighted.
        # This is probably happening from a <dfn idl-type> inside a <pre highlight=idl>.
        return None
    if attr == "nohighlight":
        return None
    elif attr == "highlight":
        return lang
    else:
        # Highlight-by-default, if applicable.
        if el.tag in ["pre", "xmp"] and hasClass(el, "idl"):
            return "webidl"
        else:
            return doc.md.defaultHighlight


def determineLineNumbers(doc, el):
    lAttr, _ = closestAttr(el, "no-line-numbers", "line-numbers")
    if lAttr == "no-line-numbers" or el.tag == "code":
        addLineNumbers = False
    elif lAttr == "line-numbers":
        addLineNumbers = True
    else:
        addLineNumbers = doc.md.lineNumbers

    lineStart = el.get("line-start")
    if lineStart is None:
        lineStart = 1
    else:
        try:
            lineStart = int(lineStart)
        except ValueError:
            die("line-start attribute must have an integer value. Got '{0}'.", lineStart, el=el)
            lineStart = 1

    lh = el.get("line-highlight")
    lineHighlights = set()
    if lh is None:
        pass
    else:
        lh = re.sub(r"\s*", "", lh)
        for item in lh.split(","):
            if "-" in item:
                # Range, format of DDD-DDD
                low,_,high = item.partition("-")
                try:
                    low = int(low)
                    high = int(high)
                except ValueError:
                    die("Error parsing line-highlight range '{0}' - must be `int-int`.", item, el=el)
                    continue
                if low >= high:
                    die("line-highlight ranges must be well-formed lo-hi - got '{0}'.", item, el=el)
                    continue
                lineHighlights.update(list(range(low, high+1)))
            else:
                try:
                    item = int(item)
                except ValueError:
                    die("Error parsing line-highlight value '{0}' - must be integers.", item, el=el)
                    continue
                lineHighlights.add(item)

    return addLineNumbers, lineStart, lineHighlights


def highlightEl(el, lang):
    text = textContent(el)
    if lang == "webidl":
        coloredText = highlightWithWebIDL(text, el=el)
    else:
        coloredText = highlightWithPygments(text, lang, el=el)
    mergeHighlighting(el, coloredText)
    addClass(el, "highlight")


def highlightWithWebIDL(text, el):
    from widlparser import parser
    '''
    Trick the widlparser emitter,
    which wants to output HTML via wrapping with start/end tags,
    into instead outputting a stack-based text format.
    A \1 indicates a new stack push;
    the text between the \1 and the \2 is the attr to be pushed.
    A \3 indicates a stack pop.
    All other text is colored with the attr currently on top of the stack.
    '''
    class IDLUI(object):
        def warn(self, msg):
            die("{0}", msg.rstrip())
    class HighlightMarker(object):
        # Just applies highlighting classes to IDL stuff.
        def markup_type_name(self, text, construct):
            return ('\1n\2', '\3')
        def markup_name(self, text, construct):
            return ('\1g\2', '\3')
        def markup_keyword(self, text, construct):
            return ('\1b\2', '\3')
        def markup_enum_value(self, text, construct):
            return ('\1s\2', '\3')

    if "\1" in text or "\2" in text or "\3" in text:
        die("WebIDL text contains some U+0001-0003 characters, which are used by the highlighter. This block can't be highlighted. :(", el=el)
        return

    widl = parser.Parser(text, IDLUI())
    return coloredTextFromWidlStack(str(widl.markup(HighlightMarker())))

def coloredTextFromWidlStack(widlText):
    coloredTexts = collections.deque()
    colors = []
    currentText = ""
    mode = "text"
    for char in widlText:
        if mode == "text":
            if char == "\1":
                if colors:
                    coloredTexts.append(ColoredText(currentText, colors[-1]))
                else:
                    coloredTexts.append(ColoredText(currentText, None))
                currentText = ""
                mode = "color"
                continue
            elif char == "\2":
                assert False, r"Encountered a \2 while in text mode"
                continue
            elif char == "\3":
                assert colors, r"Encountered a \3 without any colors on stack."
                coloredTexts.append(ColoredText(currentText, colors.pop()))
                currentText = ""
                continue
            else:
                currentText += char
        elif mode == "color":
            if char == "\1":
                assert False, r"Encountered a \1 while in color mode."
                continue
            elif char == "\2":
                colors.append(currentText)
                currentText = ""
                mode = "text"
                continue
            elif char == "\3":
                assert False, r"Encountered a \3 while in color mode."
                continue
            else:
                currentText += char
                continue
    assert len(colors) == 0, r"Colors stack wasn't empty at end, \1 and \3s aren't balanced?"
    if currentText:
        coloredTexts.append(ColoredText(currentText, None))
    return coloredTexts

def highlightWithPygments(text, lang, el):
    import pygments
    from pygments import formatters
    lexer = lexerFromLang(lang)
    if lexer is None:
        die("'{0}' isn't a known syntax-highlighting language. See http://pygments.org/docs/lexers/. Seen on:\n{1}", lang, outerHTML(el), el=el)
        return
    rawTokens = str(pygments.highlight(text, lexer, formatters.RawTokenFormatter()), encoding="utf-8")
    coloredText = coloredTextFromRawTokens(rawTokens)
    return coloredText


def mergeHighlighting(el, coloredText):
    # Merges a tree of Pygment-highlighted HTML
    # into the original element's markup.
    # This works because Pygment effectively colors each character with a highlight class,
    # merging them together into runs of text for convenience/efficiency only;
    # the markup structure is a flat list of sibling elements containing raw text
    # (and maybe some un-highlighted raw text between them).
    def createEl(color, text):
        return createElement("c-", {color:""}, text)

    def colorizeEl(el, coloredText):
        for node in childNodes(el, clear=True):
            if isElement(node):
                appendChild(el, colorizeEl(node, coloredText))
            else:
                appendChild(el, *colorizeText(node, coloredText))
        return el

    def colorizeText(text, coloredText):
        nodes = []
        while text and coloredText:
            nextColor = coloredText.popleft()
            if len(nextColor.text) <= len(text):
                if nextColor.color is None:
                    nodes.append(nextColor.text)
                else:
                    nodes.append(createEl(nextColor.color, nextColor.text))
                text = text[len(nextColor.text):]
            else:  # Need to use only part of the nextColor node
                if nextColor.color is None:
                    nodes.append(text)
                else:
                    nodes.append(createEl(nextColor.color, text))
                # Truncate the nextColor text to what's unconsumed,
                # and put it back into the deque
                nextColor = ColoredText(nextColor.text[len(text):], nextColor.color)
                coloredText.appendleft(nextColor)
                text = ''
        return nodes
    # Remove empty colored texts
    coloredText = collections.deque(x for x in coloredText if x.text)
    colorizeEl(el, coloredText)

def coloredTextFromRawTokens(text):
    colorFromName = {
        "Token.Comment": "c",
        "Token.Keyword": "k",
        "Token.Literal": "l",
        "Token.Name": "n",
        "Token.Operator": "o",
        "Token.Punctuation": "p",
        "Token.Comment.Multiline": "d",
        "Token.Comment.Preproc": "cp",
        "Token.Comment.Single": "c1",
        "Token.Comment.Special": "cs",
        "Token.Keyword.Constant": "kc",
        "Token.Keyword.Declaration": "a",
        "Token.Keyword.Namespace": "kn",
        "Token.Keyword.Pseudo": "kp",
        "Token.Keyword.Reserved": "kr",
        "Token.Keyword.Type": "b",
        "Token.Literal.Date": "ld",
        "Token.Literal.Number": "m",
        "Token.Literal.String": "s",
        "Token.Name.Attribute": "e",
        "Token.Name.Class": "nc",
        "Token.Name.Constant": "no",
        "Token.Name.Decorator": "nd",
        "Token.Name.Entity": "ni",
        "Token.Name.Exception": "ne",
        "Token.Name.Function": "nf",
        "Token.Name.Label": "nl",
        "Token.Name.Namespace": "nn",
        "Token.Name.Property": "py",
        "Token.Name.Tag": "f",
        "Token.Name.Variable": "g",
        "Token.Operator.Word": "ow",
        "Token.Literal.Number.Bin": "mb",
        "Token.Literal.Number.Float": "mf",
        "Token.Literal.Number.Hex": "mh",
        "Token.Literal.Number.Integer": "mi",
        "Token.Literal.Number.Oct": "mo",
        "Token.Literal.String.Backtick": "sb",
        "Token.Literal.String.Char": "sc",
        "Token.Literal.String.Doc": "sd",
        "Token.Literal.String.Double": "u",
        "Token.Literal.String.Escape": "se",
        "Token.Literal.String.Heredoc": "sh",
        "Token.Literal.String.Interpol": "si",
        "Token.Literal.String.Other": "sx",
        "Token.Literal.String.Regex": "sr",
        "Token.Literal.String.Single": "t",
        "Token.Literal.String.Symbol": "ss",
        "Token.Name.Variable.Class": "vc",
        "Token.Name.Variable.Global": "vg",
        "Token.Name.Variable.Instance": "vi",
        "Token.Literal.Number.Integer.Long": "il"
    }
    def addCtToList(list, ct):
        if "\n" in ct.text:
            # Break apart the formatting so that the \n is plain text,
            # so it works better with line numbers.
            textBits = ct.text.split("\n")
            list.append(ColoredText(textBits[0], ct.color))
            for bit in textBits[1:]:
                list.append(ColoredText("\n", None))
                list.append(ColoredText(bit, ct.color))
        else:
            list.append(ct)
    textList = collections.deque()
    currentCT = None
    for line in text.split("\n"):
        if not line:
            continue
        tokenName,_,tokenTextRepr = line.partition("\t")
        color = colorFromName.get(tokenName, None)
        text = eval(tokenTextRepr)
        if not text:
            continue
        if not currentCT:
            currentCT = ColoredText(text, color)
        elif currentCT.color == color:
            # Repeated color, merge into current
            currentCT = currentCT._replace(text=currentCT.text + text)
        else:
            addCtToList(textList, currentCT)
            currentCT = ColoredText(text, color)
    if currentCT:
        addCtToList(textList, currentCT)
    return textList


def normalizeLanguageName(lang):
    # Translates some names to ones Pygment understands
    if lang == "aspnet":
        return "aspx-cs"
    if lang in ["markup", "svg"]:
        return "html"
    if lang == "idl":
        return "webidl"
    return lang


def normalizeHighlightMarkers(doc):
    # Translate Prism-style highlighting into Pygment-style
    for el in findAll("[class*=language-], [class*=lang-]", doc):
        match = re.search("(?:lang|language)-(\w+)", el.get("class"))
        if match:
            el.set("highlight", match.group(1))


def lexerFromLang(lang):
    if lang in customLexers:
        return customLexers[lang]()
    try:
        from pygments.lexers import get_lexer_by_name
        return get_lexer_by_name(lang, encoding="utf-8", stripAll=True)
    except:
        return None


def addLineWrappers(el, numbers=True, start=1, highlights=None):
    # Wrap everything between each top-level newline with a line tag.
    # Add an attr for the line number, and if needed, the end line.
    if highlights is None:
        highlights = set()
    lineWrapper = E.span({"class": "line"})
    for node in childNodes(el, clear=True):
        if isElement(node):
            appendChild(lineWrapper, node)
        else:
            while True:
                if "\n" in node:
                    pre, _, post = node.partition("\n")
                    appendChild(lineWrapper, pre)
                    appendChild(el, E.span({"class": "line-no"}))
                    appendChild(el, lineWrapper)
                    lineWrapper = E.span({"class": "line"})
                    node = post
                else:
                    appendChild(lineWrapper, node)
                    break
    if len(lineWrapper):
        appendChild(el, E.span({"class": "line-no"}))
        appendChild(el, lineWrapper)
    # Number the lines
    lineNumber = start
    for lineNo, node in grouper(childNodes(el), 2):
        if numbers or lineNumber in highlights:
            lineNo.set("data-line", str(lineNumber))
        if lineNumber in highlights:
            addClass(node, "highlight-line")
            addClass(lineNo, "highlight-line")
        internalNewlines = countInternalNewlines(node)
        if internalNewlines:
            for i in range(1, internalNewlines+1):
                if (lineNumber + i) in highlights:
                    addClass(lineNo, "highlight-line")
                    addClass(node, "highlight-line")
                    lineNo.set("data-line", str(lineNumber))
            lineNumber += internalNewlines
            if numbers:
                lineNo.set("data-line-end", str(lineNumber))
        lineNumber += 1
    addClass(el, "line-numbered")
    return el

def countInternalNewlines(el):
    count = 0
    for node in childNodes(el):
        if isElement(node):
            count += countInternalNewlines(node)
        else:
            count += node.count("\n")
    return count


def getHighlightStyles():
    # To regen the styles, edit and run the below
    #from pygments import token
    #from pygments import style
    #class PrismStyle(style.Style):
    #    default_style = "#000000"
    #    styles = {
    #        token.Name: "#0077aa",
    #        token.Name.Tag: "#669900",
    #        token.Name.Builtin: "noinherit",
    #        token.Name.Variable: "#222222",
    #        token.Name.Other: "noinherit",
    #        token.Operator: "#999999",
    #        token.Punctuation: "#999999",
    #        token.Keyword: "#990055",
    #        token.Literal: "#000000",
    #        token.Literal.Number: "#000000",
    #        token.Literal.String: "#a67f59",
    #        token.Comment: "#708090"
    #    }
    #print formatters.HtmlFormatter(style=PrismStyle).get_style_defs('.highlight')
    return '''
.highlight:not(.idl) { background: hsl(24, 20%, 95%); }
code.highlight { padding: .1em; border-radius: .3em; }
pre.highlight, pre > code.highlight { display: block; padding: 1em; margin: .5em 0; overflow: auto; border-radius: 0; }
c-[a] { color: #990055 } /* Keyword.Declaration */
c-[b] { color: #990055 } /* Keyword.Type */
c-[c] { color: #708090 } /* Comment */
c-[d] { color: #708090 } /* Comment.Multiline */
c-[e] { color: #0077aa } /* Name.Attribute */
c-[f] { color: #669900 } /* Name.Tag */
c-[g] { color: #222222 } /* Name.Variable */
c-[k] { color: #990055 } /* Keyword */
c-[l] { color: #000000 } /* Literal */
c-[m] { color: #000000 } /* Literal.Number */
c-[n] { color: #0077aa } /* Name */
c-[o] { color: #999999 } /* Operator */
c-[p] { color: #999999 } /* Punctuation */
c-[s] { color: #a67f59 } /* Literal.String */
c-[t] { color: #a67f59 } /* Literal.String.Single */
c-[u] { color: #a67f59 } /* Literal.String.Double */
c-[cp] { color: #708090 } /* Comment.Preproc */
c-[c1] { color: #708090 } /* Comment.Single */
c-[cs] { color: #708090 } /* Comment.Special */
c-[kc] { color: #990055 } /* Keyword.Constant */
c-[kn] { color: #990055 } /* Keyword.Namespace */
c-[kp] { color: #990055 } /* Keyword.Pseudo */
c-[kr] { color: #990055 } /* Keyword.Reserved */
c-[ld] { color: #000000 } /* Literal.Date */
c-[nc] { color: #0077aa } /* Name.Class */
c-[no] { color: #0077aa } /* Name.Constant */
c-[nd] { color: #0077aa } /* Name.Decorator */
c-[ni] { color: #0077aa } /* Name.Entity */
c-[ne] { color: #0077aa } /* Name.Exception */
c-[nf] { color: #0077aa } /* Name.Function */
c-[nl] { color: #0077aa } /* Name.Label */
c-[nn] { color: #0077aa } /* Name.Namespace */
c-[py] { color: #0077aa } /* Name.Property */
c-[ow] { color: #999999 } /* Operator.Word */
c-[mb] { color: #000000 } /* Literal.Number.Bin */
c-[mf] { color: #000000 } /* Literal.Number.Float */
c-[mh] { color: #000000 } /* Literal.Number.Hex */
c-[mi] { color: #000000 } /* Literal.Number.Integer */
c-[mo] { color: #000000 } /* Literal.Number.Oct */
c-[sb] { color: #a67f59 } /* Literal.String.Backtick */
c-[sc] { color: #a67f59 } /* Literal.String.Char */
c-[sd] { color: #a67f59 } /* Literal.String.Doc */
c-[se] { color: #a67f59 } /* Literal.String.Escape */
c-[sh] { color: #a67f59 } /* Literal.String.Heredoc */
c-[si] { color: #a67f59 } /* Literal.String.Interpol */
c-[sx] { color: #a67f59 } /* Literal.String.Other */
c-[sr] { color: #a67f59 } /* Literal.String.Regex */
c-[ss] { color: #a67f59 } /* Literal.String.Symbol */
c-[vc] { color: #0077aa } /* Name.Variable.Class */
c-[vg] { color: #0077aa } /* Name.Variable.Global */
c-[vi] { color: #0077aa } /* Name.Variable.Instance */
c-[il] { color: #000000 } /* Literal.Number.Integer.Long */
'''

def getLineNumberStyles():
    return '''
.line-numbered {
    display: grid !important;
    grid-template-columns: min-content 1fr;
    grid-auto-flow: row;
}
.line-numbered > *,
.line-numbered::before,
.line-numbered::after {
    grid-column: 1/-1;
}
.line-no {
    grid-column: 1;
    color: gray;
}
.line {
    grid-column: 2;
}
.line:hover {
    background: rgba(0,0,0,.05);
}
.line-no[data-line]::before {
    padding: 0 .5em 0 .1em;
    content: attr(data-line);
}
.line-no[data-line-end]::after {
    padding: 0 .5em 0 .1em;
    content: attr(data-line-end);
}
'''

def getLineHighlightingStyles():
    return '''
.line-numbered {
    display: grid !important;
    grid-template-columns: min-content 1fr;
    grid-auto-flow: rows;
}
.line-numbered > *,
.line-numbered::before,
.line-numbered::after {
    grid-column: 1/-1;
}
.line-no {
    grid-column: 1;
    color: gray;
}
.line {
    grid-column: 2;
}
.line.highlight-line {
    background: rgba(0,0,0,.05);
}
.line-no.highlight-line {
    background: rgba(0,0,0,.05);
    color: #444;
    font-weight: bold;
}
.line-no.highlight-line[data-line]::before {
    padding: 0 .5em 0 .1em;
    content: attr(data-line);
}
.line-no.highlight-line[data-line-end]::after {
    padding: 0 .5em 0 .1em;
    content: attr(data-line-end);
}
'''


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.zip_longest(fillvalue=fillvalue, *args)



# -*- coding: utf-8 -*-


import io

from .. import config
from ..html import findAll, textContent, removeNode, E, addClass, appendChild, clearContents
from ..messages import *

def processWptElements(doc):
	pathPrefix = doc.md.wptPathPrefix

	atLeastOneElement = False
	testData = None
	# <wpt> elements
	wptElements = findAll("wpt", doc)
	seenTestNames = set()
	for el in wptElements:
		atLeastOneElement = True
		if testData is None:
			testData = loadTestData(doc)
		testNames = testNamesFromEl(el, pathPrefix=pathPrefix)
		for testName in testNames:
			if testName not in testData:
				die("Couldn't find WPT test '{0}' - did you misspell something?", testName, el=el)
				continue
			seenTestNames.add(testName)
		createHTML(doc, el, testNames)

	# <wpt-rest> elements
	wptRestElements = findAll("wpt-rest", doc)
	if wptRestElements and testData is None:
		testData = loadTestData(doc)
	if len(wptRestElements) > 1:
		die("Only one <wpt-rest> element allowed per document, you have {0}.", len(wptRestElements))
		wptRestElements = wptRestElements[0:1]
	elif len(wptRestElements) == 1:
		localPrefix = wptRestElements[0].get("pathprefix")
		if localPrefix is not None:
			pathPrefix = localPrefix
		if pathPrefix is None:
			die("Can't use <wpt-rest> without either a pathprefix="" attribute or a 'WPT Path Prefix' metadata.")
			return
		atLeastOneElement = True
		prefixedNames = [p for p in testData if prefixInPath(pathPrefix, p) and p not in seenTestNames]
		if len(prefixedNames) == 0:
			die("Couldn't find any tests with the path prefix '{0}'.", pathPrefix)
			return
		createHTML(doc, wptRestElements[0], prefixedNames)
		warn("<wpt-rest> is intended for debugging only. Move the tests to <wpt> elements next to what they're testing.")
	else:
		if pathPrefix:
			if testData is None:
				testData = loadTestData(doc)
			checkForOmittedTests(pathPrefix, testData, seenTestNames)

	if atLeastOneElement:
		doc.extraStyles['style-wpt'] = wptStyle



def createHTML(doc, blockEl, testNames):
	if doc.md.wptDisplay == "none":
		removeNode(blockEl)
	elif doc.md.wptDisplay == "inline":
		blockEl.tag = "ul"
		addClass(blockEl, "wpt-tests-block")
		clearContents(blockEl)
		for testName in testNames:
			if ".https." in testName or ".serviceworker." in testName:
				liveTestScheme = "https"
			else:
				liveTestScheme = "http"
			_,_,lastNameFragment = testName.rpartition("/")
			singleTestEl = E.li({"class": "wpt-test"},
				E.a({"href": "https://wpt.fyi/results/"+testName}, lastNameFragment),
				" ",
				E.a({"title": testName, "href": "{0}://web-platform-tests.live/{1}".format(liveTestScheme, testName)}, E.small("(live test)")),
				" ",
				E.a({"href": "https://github.com/web-platform-tests/wpt/blob/master/"+testName}, E.small("(source)")))
			appendChild(blockEl, singleTestEl)
	else:
		die("Programming error, uncaught WPT Display value in createHTML.")


def testNamesFromEl(el, pathPrefix=None):
	testNames = []
	localPrefix = el.get("pathprefix")
	if localPrefix is not None:
		pathPrefix = localPrefix
	for name in [x.strip() for x in textContent(el).split("\n")]:
		if name == "":
			continue
		testName = prefixPlusPath(pathPrefix, name)
		testNames.append(testName)
	return testNames


def prefixPlusPath(prefix, path):
	# Join prefix to path, normalizing slashes
	if path.startswith("/"):
		return path[1:]
	prefix = normalizePathSegment(prefix)
	if prefix is None:
		return path
	return prefix + path

def prefixInPath(prefix, path):
	if prefix is None:
		return False
	return path.startswith(normalizePathSegment(prefix))

def normalizePathSegment(pathSeg):
	# No slash at front, yes slash at end
	if pathSeg is None:
		return None
	if pathSeg.startswith("/"):
		pathSeg = pathSeg[1:]
	if not pathSeg.endswith("/"):
		pathSeg += "/"
	return pathSeg

def checkForOmittedTests(pathPrefix, testData, seenTestNames):
	unseenTests = []
	for testPath in testData.keys():
		if ".tentative." in testPath:
			continue
		if prefixInPath(pathPrefix, testPath):
			if testPath not in seenTestNames:
				unseenTests.append(testPath)
	if unseenTests:
		warn("There are {0} WPT tests underneath your path prefix that aren't in your document and must be added:\n{1}",
			len(unseenTests),
			"\n".join("  " + path for path in sorted(unseenTests)))


def loadTestData(doc):
	paths = {}
	for line in doc.dataFile.fetch("wpt-tests.txt", str=True).split("\n")[1:]:
		testType,_,testPath = line.strip().partition(" ")
		paths[testPath] = testType
	return paths

def xor(a, b):
	return bool(a) != bool(b)

wptStyle = '''
.wpt-tests-block {
	list-style: none;
	border-left: .5em solid hsl(290, 70%, 60%);
	background: hsl(290, 70%, 95%);
	margin: 1em auto;
	padding: .5em;
	display: grid;
	grid-template-columns: 1fr auto auto;
	grid-column-gap: .5em;
}
.wpt-tests-block::before {
	content: "Tests";
	grid-column: 1/-1;
	color: hsl(290, 70%, 30%);
	text-transform: uppercase;
}
.wpt-test {
	display: contents;
}
.wpt-test > a {
	text-decoration: underline;
	border: none;
}
'''



# -*- coding: utf-8 -*-

import re
from itertools import *

from .. import Line
from ..html import escapeAttr
from ..messages import *


def parse(lines, numSpacesForIndentation, features=None, opaqueElements=None, blockElements=None):
    fromStrings = False
    if any(isinstance(l, str) for l in lines):
        fromStrings = True
        lines = [Line.Line(-1, l) for l in lines]
    lines = Line.rectify(lines)
    tokens = tokenizeLines(lines, numSpacesForIndentation, features, opaqueElements=opaqueElements, blockElements=blockElements)
    html = parseTokens(tokens, numSpacesForIndentation)
    if fromStrings:
        return [l.text for l in html]
    else:
        return html


def tokenizeLines(lines, numSpacesForIndentation, features=None, opaqueElements=None, blockElements=None):
    # Turns lines of text into block tokens,
    # which'll be turned into MD blocks later.
    # Every token *must* have 'type', 'raw', and 'prefix' keys.

    if features is None:
        features = {"headings"}

    # Inline elements that are allowed to start a "normal" line of text.
    # Any other element will instead be an HTML line and will close paragraphs, etc.
    inlineElements = {"a", "em", "strong", "small", "s", "cite", "q", "dfn", "abbr", "data", "time", "code", "var", "samp", "kbd", "sub", "sup", "i", "b", "u", "mark", "ruby", "bdi", "bdo", "span", "br", "wbr", "img", "meter", "progress", "css", "l"}
    if blockElements is None:
        blockElements = []

    def inlineElementStart(line):
        # Whether or not the line starts with an inline element
        match = re.match(r"\s*</?([\w-]+)", line)
        if not match:
            return True
        tagname = match.group(1)
        if tagname in inlineElements:
            return True
        if "-" in tagname and tagname not in blockElements:
            # Assume custom elements are inline by default
            return True
        return False

    tokens = []
    rawStack = []
    if opaqueElements is None:
        opaqueElements = []
    opaqueElements += ["pre", "xmp", "script", "style"]
    rawElements = "|".join(re.escape(x) for x in opaqueElements)

    for l in lines:

        # Three kinds of "raw" elements, which prevent markdown processing inside of them.
        # 1. <pre> and manual opaque elements, which can contain markup and so can nest.
        # 2. <xmp>, <script>, and <style>, which contain raw text, can't nest.
        # 3. Markdown code blocks, which contain raw text, can't nest.
        #
        # The rawStack holds tokens like
        # {"type":"fenced", "tag":"````", "nest":False}

        if rawStack:
            # Inside at least one raw element that turns off markdown.
            # First see if this line will *end* the raw element.
            endTag = rawStack[-1]
            if endTag['type'] == "element" and re.search(endTag['tag'], l.text):
                rawStack.pop()
                tokens.append({'type':'raw', 'prefixlen':float('inf'), 'line':l})
                continue
            elif endTag['type'] == "fenced" and re.match(r"\s*{0}{1}*\s*$".format(endTag['tag'], endTag['tag'][0]), l.text):
                rawStack.pop()
                l.text = "</xmp>"
                tokens.append({'type':'raw', 'prefixlen':float('inf'), 'line':l})
                continue
            elif not endTag['nest']:
                # Just an internal line, but for the no-nesting elements,
                # so guaranteed no more work needs to be done.
                tokens.append({'type':'raw', 'prefixlen':float('inf'), 'line':l})
                continue

        # We're either in a nesting raw element or not in a raw element at all,
        # so check if the line starts a new element.
        match = re.match(r"(\s*)(`{3,}|~{3,})([^`]*)$", l.text)
        if match:
            ws,tag,infoString = match.groups()
            rawStack.append({"type":"fenced", "tag":tag, "nest":False})
            infoString = infoString.strip()
            if infoString:
                # For now, I only care about lang
                lang = infoString.split(" ")[0]
                classAttr = " class='language-{0}'".format(escapeAttr(lang))
            else:
                classAttr = ""
            l.text = '{0}<xmp{1}>'.format(ws, classAttr)
            tokens.append({'type':'raw', 'prefixlen':prefixLen(ws, numSpacesForIndentation), 'line':l})
            continue
        match = re.match(r"\s*<({0})[ >]".format(rawElements), l.text)
        if match:
            tokens.append({'type':'raw', 'prefixlen':prefixLen(l.text, numSpacesForIndentation), 'line':l})
            if re.search(r"</({0})>".format(match.group(1)), l.text):
                # Element started and ended on same line, cool, don't need to do anything.
                pass
            else:
                nest = match.group(1) not in ["xmp", "script", "style"]
                rawStack.append({'type':'element', 'tag':"</{0}>".format(match.group(1)), 'nest':nest})
            continue
        if rawStack:
            tokens.append({'type':'raw', 'prefixlen':float('inf'), 'line':l})
            continue

        line = l.text.strip()

        if line == "":
            token = {'type':'blank',}
        # FIXME: Detect the heading ID from heading lines
        elif "headings" in features and re.match(r"={3,}\s*$", line):
            # h1 underline
            match = re.match(r"={3,}\s$", line)
            token = {'type':'equals-line'}
        elif "headings" in features and re.match(r"-{3,}\s*$", line):
            # h2 underline
            match = re.match(r"-{3,}\s*$", line)
            token = {'type':'dash-line'}
        elif "headings" in features and re.match(r"(#{1,5})\s+(.+?)(\1\s*\{#[^ }]+\})?\s*$", line):
            # single-line heading
            match = re.match(r"(#{1,5})\s+(.+?)(\1\s*\{#[^ }]+\})?\s*$", line)
            level = len(match.group(1)) + 1
            token = {'type':'heading', 'text': match.group(2).strip(), 'level': level}
            match = re.search(r"\{#([^ }]+)\}\s*$", line)
            if match:
                token['id'] = match.group(1)
        elif re.match(r"((\*\s*){3,})$|((-\s*){3,})$|((_\s*){3,})$", line):
            token = {'type':'rule'}
        elif re.match(r"-?\d+\.\s", line):
            match = re.match(r"(-?\d+)\.\s+(.*)", line)
            token = {'type':'numbered', 'text': match.group(2), 'num': int(match.group(1))}
        elif re.match(r"-?\d+\.$", line):
            token = {'type':'numbered', 'text': "", 'num': int(line[:-1])}
        elif re.match(r"[*+-]\s", line):
            match = re.match(r"[*+-]\s+(.*)", line)
            token = {'type':'bulleted', 'text': match.group(1)}
        elif re.match(r"[*+-]$", line):
            token = {'type':'bulleted', 'text': ""}
        elif re.match(r":{1,2}\s+", line):
            match = re.match(r"(:{1,2})\s+(.*)", line)
            type = 'dt' if len(match.group(1)) == 1 else 'dd'
            token = {'type':type, 'text': match.group(2)}
        elif re.match(r":{1,2}$", line):
            match = re.match(r"(:{1,2})", line)
            type = 'dt' if len(match.group(1)) == 1 else 'dd'
            token = {'type':type, 'text': ""}
        elif re.match(r">", line):
            match = re.match(r">\s?(.*)", line)
            token = {'type':'blockquote', 'text':match.group(1)}
        elif re.match(r"<", line):
            if re.match(r"<<|<\{", line) or inlineElementStart(line):
                token = {'type':'text', 'text': line}
            else:
                token = {'type':'htmlblock'}
        else:
            token = {'type':'text', 'text': line}

        if token['type'] == "blank":
            token['prefixlen'] = float('inf')
        else:
            token['prefixlen'] = prefixLen(l.text, numSpacesForIndentation)
        token['line'] = l
        tokens.append(token)

    #for token in tokens:
    #    print (" " * (11 - len(token['type']))) + token['type'] + ": " + token['line'].text.rstrip()

    return tokens


def stripComments(lines):
    '''
    Eagerly strip comments, because the serializer can't output them right now anyway,
    and they trigger some funky errors.
    '''
    output = []
    inComment = False
    wholeCommentLines = 0
    for line in lines:
        text, inComment = stripCommentsFromLine(line.text, inComment)
        if (line.text != text and text.strip() == "") or (line.text.strip() == "" and inComment):
            # First covers the entire line being stripped away by comment-removal.
            # Second covers an empty line that was fully inside a comment.
            # (If a comment started or ended on that line, it wouldn't start out empty.)
            # By removing these entirely, we avoid breaking Markdown constructs with their middles commented out or something.
            # (Rather than leaving them in as blank lines.)
            continue
        else:
            # Otherwise, just process whatever's left as normal.
            if line.text.endswith("\n") and not text.endswith("\n"):
                # Put back the newline, in case it got swallowed by an unclosed comment.
                text += "\n"
            line.text = text
            output.append(line)
    return output


def stripCommentsFromLine(line, inComment=False):
    # Removes HTML comments from the line.
    # Returns true if the comment wasn't closed by the end of the line
    if inComment:
        # The line starts out in a comment.
        pre,sep,post = line.partition("-->")
        if sep == "":
            # The entire line is a comment
            return "", True
        else:
            # Drop the comment part, see if there are any more
            return stripCommentsFromLine(post)
    else:
        # The line starts out as non-comment content.
        pre,sep,post = line.partition("<!--")
        if sep == "":
            # No comments in the line
            return pre, False
        else:
            # Keep the non-comment part, see if there's any more to do
            res,inComment = stripCommentsFromLine(post, inComment=True)
            return pre + res, inComment


def prefixLen(text, numSpacesForIndentation):
    i = 0
    prefixLen = 0
    while i < len(text):
        if text[i] == "\t":
            i += 1
            prefixLen += 1
        elif text[i:i + numSpacesForIndentation] == " " * numSpacesForIndentation:
            i += numSpacesForIndentation
            prefixLen += 1
        else:
            break
    return prefixLen


def stripPrefix(token, numSpacesForIndentation, len):
    '''Removes len number of prefix groups'''

    text = token['line'].text

    # Don't mess with "infinite" prefix lines
    if token['prefixlen'] == float('inf'):
        return text

    # Allow empty lines
    if text.strip() == "":
        return text

    offset = 0
    for x in range(len):
        if text[offset] == "\t":
            offset += 1
        elif text[offset:offset + numSpacesForIndentation] == " " * numSpacesForIndentation:
            offset += numSpacesForIndentation
        else:
            die("Line {i} isn't indented enough (needs {0} indent{plural}) to be valid Markdown:\n\"{1}\"", len, text[:-1], plural="" if len == 1 else "s", i=token['line'].i)
    return text[offset:]


def parseTokens(tokens, numSpacesForIndentation):
    '''
    Token types:
    eof
    blank
    equals-line
    dash-line
    heading
    rule
    numbered
    bulleted
    dt
    dd
    text
    htmlblock
    blockquote
    raw
    '''
    stream = TokenStream(tokens, numSpacesForIndentation)
    lines = []

    while True:
        if stream.ended():
            break
        elif stream.currtype() in ('raw', 'htmlblock'):
            lines.append(stream.currline())
            stream.advance()
        elif stream.currtype() == 'heading':
            lines += parseSingleLineHeading(stream)
        elif stream.currtype() == 'text' and stream.nexttype() in ('equals-line', 'dash-line'):
            lines += parseMultiLineHeading(stream)
        elif stream.currtype() == 'text' and stream.prevtype() == 'blank':
            lines += parseParagraph(stream)
        elif stream.currtype() == 'rule' or stream.currtype() == 'dash-line':
            lines += parseHorizontalRule(stream)
        elif stream.currtype() == 'bulleted':
            lines += parseBulleted(stream)
        elif stream.currtype() == 'numbered':
            lines += parseNumbered(stream, start=stream.currnum())
        elif stream.currtype() in ("dt", "dd"):
            lines += parseDl(stream)
        elif stream.currtype() == "blockquote":
            lines += parseBlockquote(stream)
        else:
            lines.append(stream.currline())
            stream.advance()

    #for line in lines:
    #    print "«{0}»".format(line.text.rstrip())

    return lines

def lineFromStream(stream, text):
    # Shortcut for when you're producing a new line from the currline in the stream,
    # with some modified text.
    return Line.Line(stream.currline().i,  text)

# Each parser gets passed the stream
# and must return the lines it returns.
# The stream should be advanced to the *next* line,
# after the lines you've used up dealing with the construct.


def parseSingleLineHeading(stream):
    if "id" in stream.curr():
        idattr = " id='{0}'".format(stream.currid())
    else:
        idattr = ""
    lines = [lineFromStream(stream, "<h{level}{idattr} line-number={i}>{text}</h{level}>\n".format(idattr=idattr, i=stream.currline().i, **stream.curr()))]
    stream.advance()
    return lines


def parseMultiLineHeading(stream):
    if stream.nexttype() == "equals-line":
        level = 2
    elif stream.nexttype() == "dash-line":
        level = 3
    else:
        die("Markdown parser error: tried to parse a multiline heading from:\n{0}{1}{2}", stream.prevraw(), stream.currraw(), stream.nextraw())
    match = re.search(r"(.*?)\s*\{\s*#([^ }]+)\s*\}\s*$", stream.currtext())
    if match:
        text = match.group(1)
        idattr = "id='{0}'".format(match.group(2))
    else:
        text = stream.currtext()
        idattr = ""
    lines = [lineFromStream(stream,  "<h{level} {idattr} line-number={i}>{htext}</h{level}>\n".format(idattr=idattr, level=level, htext=text, i=stream.currline().i, **stream.curr()))]
    stream.advance(2)
    return lines


def parseHorizontalRule(stream):
    lines = [lineFromStream(stream,  "<hr>\n")]
    stream.advance()
    return lines


def parseParagraph(stream):
    line = stream.currtext()
    i = stream.currline().i
    initialPrefixLen = stream.currprefixlen()
    endTag = "</p>"
    if re.match(r"note[:,]\s*", line, re.I):
        p = "<p class='replace-with-note-class'>"
        matchNote = re.match(r"(note[:,])(\s*)(.*)", line, re.I)
        if matchNote:
            line = matchNote.group(3)
            p += "<span>{note}</span>{ws}".format(note=matchNote.group(1), ws=matchNote.group(2))
    elif line.lower().startswith("issue:"):
        line = line[6:]
        p = "<p line-number={0} class='replace-with-issue-class'>".format(i)
    elif line.lower().startswith("assert:"):
        p = "<p line-number={0} class='replace-with-assertion-class'>".format(i)
    elif line.lower().startswith("advisement:"):
        line = line[11:]
        p = "<p line-number={0}><strong class='replace-with-advisement-class'>".format(i)
        endTag = "</strong></p>"
    else:
        match = re.match(r"issue\(([^)]+)\):(.*)", line, re.I)
        if match:
            line = match.group(2)
            p = "<p line-number={0} data-remote-issue-id='{1}' class='replace-with-issue-class'>".format(i, match.group(1))
        else:
            p = "<p line-number={0}>".format(i)
    lines = [lineFromStream(stream, "{0}{1}\n".format(p, line))]
    while True:
        stream.advance()
        if stream.currtype() not in ["text"] or stream.currprefixlen() < initialPrefixLen:
            lines[-1].text = lines[-1].text.rstrip() + endTag + "\n"
            return lines
        lines.append(stream.currline())


def parseBulleted(stream):
    prefixLen = stream.currprefixlen()
    numSpacesForIndentation = stream.numSpacesForIndentation
    ul_i = stream.currline().i

    def parseItem(stream):
        # Assumes it's being called with curr being a bulleted line.
        # Remove the bulleted part from the line
        firstLine = stream.currtext() + "\n"
        i = stream.currline().i
        lines = [lineFromStream(stream,  firstLine)]
        while True:
            stream.advance()
            # All the conditions that indicate we're *past* the end of the item.
            if stream.currtype() == 'bulleted' and stream.currprefixlen() == prefixLen:
                return lines, i
            if stream.currprefixlen() < prefixLen:
                return lines, i
            if stream.currtype() == 'blank' and stream.nexttype() != 'bulleted' and stream.nextprefixlen() <= prefixLen:
                return lines, i
            if stream.currtype() == 'eof':
                return lines, i
            # Remove the prefix from each line before adding it.
            lines.append(lineFromStream(stream,  stripPrefix(stream.curr(), numSpacesForIndentation, prefixLen + 1)))

    def getItems(stream):
        while True:
            # The conditions that indicate we're past the end of the list itself
            if stream.currtype() == 'eof':
                return
            if stream.currtype() == 'blank' and stream.nexttype() != 'bulleted' and stream.nextprefixlen() <= prefixLen:
                return
            if stream.currprefixlen() < prefixLen:
                return
            if stream.currtype() == 'blank':
                stream.advance()
            yield parseItem(stream)

    lines = [Line.Line(-1,  "<ul data-md line-number={0}>".format(ul_i))]
    for li_lines, i in getItems(stream):
        lines.append(Line.Line(-1, "<li data-md line-number={0}>".format(i)))
        lines.extend(parse(li_lines, numSpacesForIndentation))
        lines.append(Line.Line(-1, "</li>"))
    lines.append(Line.Line(-1, "</ul>"))
    return lines


def parseNumbered(stream, start=1):
    prefixLen = stream.currprefixlen()
    ol_i = stream.currline().i
    numSpacesForIndentation = stream.numSpacesForIndentation

    def parseItem(stream):
        # Assumes it's being called with curr being a numbered line.
        # Remove the numbered part from the line
        firstLine = stream.currtext() + "\n"
        i = stream.currline().i
        lines = [lineFromStream(stream, firstLine)]
        while True:
            stream.advance()
            # All the conditions that indicate we're *past* the end of the item.
            if stream.currtype() == 'numbered' and stream.currprefixlen() == prefixLen:
                return lines,i
            if stream.currprefixlen() < prefixLen:
                return lines,i
            if stream.currtype() == 'blank' and stream.nexttype() != 'numbered' and stream.nextprefixlen() <= prefixLen:
                return lines,i
            if stream.currtype() == 'eof':
                return lines,i
            # Remove the prefix from each line before adding it.
            lines.append(lineFromStream(stream, stripPrefix(stream.curr(), numSpacesForIndentation, prefixLen + 1)))

    def getItems(stream):
        while True:
            # The conditions that indicate we're past the end of the list itself
            if stream.currtype() == 'eof':
                return
            if stream.currtype() == 'blank' and stream.nexttype() != 'numbered' and stream.nextprefixlen() <= prefixLen:
                return
            if stream.currprefixlen() < prefixLen:
                return
            if stream.currtype() == 'blank':
                stream.advance()
            yield parseItem(stream)

    if start == 1:
        lines = [Line.Line(-1, "<ol data-md line-number={0}>".format(ol_i))]
    else:
        lines = [Line.Line(-1, "<ol data-md start='{0}' line-number={1}>".format(start, ol_i))]
    for li_lines,i in getItems(stream):
        lines.append(Line.Line(-1, "<li data-md line-number={0}>".format(i)))
        lines.extend(parse(li_lines, numSpacesForIndentation))
        lines.append(Line.Line(-1, "</li>"))
    lines.append(Line.Line(-1, "</ol>"))
    return lines


def parseDl(stream):
    prefixLen = stream.currprefixlen()
    dl_i = stream.currline().i
    numSpacesForIndentation = stream.numSpacesForIndentation

    def parseItem(stream):
        # Assumes it's being called with curr being a :/:: prefixed line.
        firstLine = stream.currtext() + "\n"
        i = stream.currline().i
        type = stream.currtype()
        lines = [lineFromStream(stream, firstLine)]
        while True:
            stream.advance()
            # All the conditions that indicate we're *past* the end of the item.
            if stream.currtype() in ('dt', 'dd') and stream.currprefixlen() == prefixLen:
                return type, lines, i
            if stream.currprefixlen() < prefixLen:
                return type, lines, i
            if stream.currtype() == 'blank' and stream.nexttype() not in ('dt', 'dd') and stream.nextprefixlen() <= prefixLen:
                return type, lines, i
            if stream.currtype() == 'eof':
                return type, lines, i
            # Remove the prefix from each line before adding it.
            lines.append(lineFromStream(stream, stripPrefix(stream.curr(), numSpacesForIndentation, prefixLen + 1)))

    def getItems(stream):
        while True:
            # The conditions that indicate we're past the end of the list itself
            if stream.currtype() == 'eof':
                return
            if stream.currtype() == 'blank' and stream.nexttype() not in ('dt', 'dd') and stream.nextprefixlen() <= prefixLen:
                return
            if stream.currprefixlen() < prefixLen:
                return
            if stream.currtype() == 'blank':
                stream.advance()
            yield parseItem(stream)

    lines = [Line.Line(-1, "<dl data-md line-number={0}>".format(dl_i))]
    for type, di_lines, i in getItems(stream):
        lines.append(Line.Line(-1, "<{0} data-md line-number={1}>".format(type, i)))
        lines.extend(parse(di_lines, numSpacesForIndentation))
        lines.append(Line.Line(-1, "</{0}>".format(type)))
    lines.append(Line.Line(-1, "</dl>"))
    return lines


def parseBlockquote(stream):
    prefixLen = stream.currprefixlen()
    i = stream.currline().i
    lines = [lineFromStream(stream, stream.currtext()+"\n")]
    while True:
        stream.advance()
        if stream.currprefixlen() < prefixLen:
            break
        if stream.currtype() in ["blockquote", "text"]:
            lines.append(lineFromStream(stream, stream.currtext()+"\n"))
        else:
            break
    return [Line.Line(-1, "<blockquote line-number={0}>\n".format(i))] + parse(lines, stream.numSpacesForIndentation) + [Line.Line(-1, "</blockquote>\n")]


class TokenStream:
    def __init__(self, tokens, numSpacesForIndentation, before={'type':'blank','prefixlen':0}, after={'type':'eof','prefixlen':0}):
        self.tokens = tokens
        self.i = 0
        self.numSpacesForIndentation = numSpacesForIndentation
        self.before = before
        self.after = after

    def __len__(self):
        return len(self.tokens)

    def ended(self):
        return self.i >= len(self)

    def nth(self, i):
        if i < 0:
            return self.before
        elif i >= len(self):
            return self.after
        else:
            return self.tokens[i]

    def prev(self, i=1):
        return self.nth(self.i - i)

    def curr(self):
        return self.nth(self.i)

    def next(self, i=1):
        return self.nth(self.i + i)

    def advance(self, i=1):
        self.i += i
        return self.curr()

    def __getattr__(self, name):
        if len(name) >= 5 and name[0:4] in ("prev", "curr", "next"):
            tokenDir = name[0:4]
            attrName = name[4:]

            def _missing(i=1):
                if tokenDir == "prev":
                    tok = self.prev(i)
                elif tokenDir == "next":
                    tok = self.next(i)
                else:
                    tok = self.curr()
                if attrName in tok:
                    return tok[attrName]
                else:
                    raise AttributeError(attrName)
                    return tok['raw']
            return _missing



# -*- coding: utf-8 -*-

import sys
from collections import Counter
import lxml.html
from . import constants



messages = set()
messageCounts = Counter()


def p(msg):
    if isinstance(msg, tuple):
        msg, ascii = msg
    else:
        ascii = None
    if constants.quiet == float("infinity"):
        return
    try:
        print(msg)
    except UnicodeEncodeError:
        if ascii is not None:
            print(ascii.encode("ascii", "replace"))
        else:
            warning = formatMessage("warning", "Your console does not understand Unicode.\n  Messages may be slightly corrupted.")
            if warning not in messages:
                print(warning)
                messages.add(warning)
            print(msg.encode("ascii", "xmlcharrefreplace"))


def die(msg, *formatArgs, **namedArgs):
    lineNum = None
    if 'el' in namedArgs and namedArgs['el'].get("line-number"):
        lineNum = namedArgs['el'].get("line-number")
    elif namedArgs.get("lineNum", None):
        lineNum = namedArgs['lineNum']
    msg = formatMessage("fatal", msg.format(*formatArgs, **namedArgs), lineNum=lineNum)
    if msg not in messages:
        messageCounts["fatal"] += 1
        messages.add(msg)
        if constants.quiet < 3:
            p(msg)
    if constants.errorLevelAt("fatal"):
        errorAndExit()


def linkerror(msg, *formatArgs, **namedArgs):
    lineNum = None
    suffix = ""
    if 'el' in namedArgs:
        el = namedArgs['el']
        if el.get("line-number"):
            lineNum = el.get("line-number")
        else:
            if el.get("bs-autolink-syntax"):
                suffix = "\n{0}".format(el.get("bs-autolink-syntax"))
            else:
                suffix = "\n{0}".format(lxml.html.tostring(namedArgs['el'], with_tail=False, encoding="unicode"))
    elif namedArgs.get("lineNum", None):
        lineNum = namedArgs['lineNum']
    msg = formatMessage("link", msg.format(*formatArgs, **namedArgs)+suffix, lineNum=lineNum)
    if msg not in messages:
        messageCounts["linkerror"] += 1
        messages.add(msg)
        if constants.quiet < 2:
                p(msg)
    if constants.errorLevelAt("link-error"):
        errorAndExit()


def warn(msg, *formatArgs, **namedArgs):
    lineNum = None
    if 'el' in namedArgs and namedArgs['el'].get("line-number"):
        lineNum = namedArgs['el'].get("line-number")
    elif namedArgs.get("lineNum", None):
        lineNum = namedArgs['lineNum']
    msg = formatMessage("warning", msg.format(*formatArgs, **namedArgs), lineNum=lineNum)
    if msg not in messages:
        messageCounts["warning"] += 1
        messages.add(msg)
        if constants.quiet < 1:
                p(msg)
    if constants.errorLevelAt("warning"):
        errorAndExit()


def say(msg, *formatArgs, **namedArgs):
    if constants.quiet < 1:
        p(formatMessage("message", msg.format(*formatArgs, **namedArgs)))


def success(msg, *formatArgs, **namedArgs):
    if constants.quiet < 4:
        msg = formatMessage("success", msg.format(*formatArgs, **namedArgs))
        p(msg)


def failure(msg, *formatArgs, **namedArgs):
    if constants.quiet < 4:
        msg = formatMessage("failure", msg.format(*formatArgs, **namedArgs))
        p(msg)


def resetSeenMessages():
    global messages
    messages = set()
    global messageCounts
    messageCounts = Counter()


def printColor(text, color="white", *styles):
    if constants.printMode == "console":
        colorsConverter = {
            "black": 30,
            "red": 31,
            "green": 32,
            "yellow": 33,
            "blue": 34,
            "magenta": 35,
            "cyan": 36,
            "light gray": 37,
            "dark gray": 90,
            "light red": 91,
            "light green": 92,
            "light yellow": 93,
            "light blue": 94,
            "light magenta": 95,
            "light cyan": 96,
            "white": 97
        }
        stylesConverter = {
            "normal": 0,
            "bold": 1,
            "bright": 1,
            "dim": 2,
            "underline": 4,
            "underlined": 4,
            "blink": 5,
            "reverse": 7,
            "invert": 7,
            "hidden": 8
        }

        colorNum = colorsConverter[color.lower()]
        styleNum = ";".join(str(stylesConverter[style.lower()]) for style in styles)
        return "\033[{0};{1}m{text}\033[0m".format(styleNum, colorNum, text=text)
    else:
        return text


def formatMessage(type, text, lineNum=None):
    if constants.printMode == "markup":
        text = text.replace("<", "&lt;")
        if type == "fatal":
            return "<fatal>{0}</fatal>".format(text)
        elif type == "link":
            return "<linkerror>{0}</linkerror>".format(text)
        elif type == "warning":
            return "<warning>{0}</warning>".format(text)
        elif type == "message":
            return "<message>{0}</message>".format(text)
        elif type == "success":
            return "<final-success>{0}</final-success>".format(text)
        elif type == "failure":
            return "<final-failure>{0}</final-failure>".format(text)
    else:
        if type == "message":
            return text
        if type == "success":
            return (printColor(" ✔ ", "green", "invert") + " " + text, printColor("YAY", "green", "invert") + " " + text)
        if type == "failure":
            return (printColor(" ✘ ", "red", "invert") + " " + text, printColor("ERR", "red", "invert") + " " + text)
        if type == "fatal":
            headingText = "FATAL ERROR"
            color = "red"
        elif type == "link":
            headingText = "LINK ERROR"
            color = "yellow"
        elif type == "warning":
            headingText = "WARNING"
            color = "light cyan"
        if lineNum is not None:
            headingText = "LINE {0}".format(lineNum)
        return printColor(headingText + ":", color, "bold") + " " + text

def errorAndExit():
    failure("Did not generate, due to fatal errors")
    sys.exit(2)



from collections import OrderedDict, Callable


class DefaultOrderedDict(OrderedDict):
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
                not isinstance(default_factory, Callable)):
            raise TypeError('first argument must be callable')
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, list(self.items())

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(list(self.items())))

    def __repr__(self):
        return 'OrderedDefaultDict(%s, %s)' % (self.default_factory,
                                               OrderedDict.__repr__(self))



# -*- coding: utf-8 -*-


class Nil(object):
    '''Super-None, falsey and returns itself from every method/attribute/etc'''

    def __repr__(self):
        return "Nil()"

    def __str__(self):
        return "Nil"

    def __bool__(self):
        return False

    def __call__(self, *args, **kwargs):
        return self

    def __getitem__(self, key):
        return self

    def __setitem__(self, key, val):
        return self

    def __delitem__(self, key, val):
        return self

    def __getattr__(self, name):
        return self

    def __setattr__(self, name, value):
        return self

    def __delattr__(self, name, value):
        return self

    def __eq__(self, other):
        if isinstance(other, Nil) or other is None:
            return True
        return False

    def __iter__(self):
        return iter([])


# -*- coding: utf-8 -*-
import io
import json
import os
from collections import OrderedDict
from . import config
from .html import *  # noqa
from .messages import *


def addMdnPanels(doc):
    if not doc.md.includeMdnPanels:
        return

    try:
        filename = doc.md.vshortname+".json"
        datafile = doc.dataFile.fetch("mdn", filename, str=True)
    except IOError:
        try:
            filename = doc.md.shortname+".json"
            datafile = doc.dataFile.fetch("mdn", filename, str=True)
        except IOError:
            if doc.md.includeMdnPanels == "maybe":
                # if "maybe", failure is fine, don't complain
                pass
            else:
                die(f"Couldn't find the MDN data for '{doc.md.vshortname}' nor '{doc.md.shortname}'.")
            return
    try:
        data = json.loads(datafile, object_pairs_hook=OrderedDict)
    except Exception as e:
        die(f"Couldn't load MDN Spec Links data for this spec.\n{e}")
        return

    panels = panelsFromData(doc, data)
    if panels:
        doc.extraScripts["script-mdn-anno"] = '''
            document.body.addEventListener("click", (e) => {
                if(e.target.closest(".mdn-anno-btn")) {
                    e.target.closest(".mdn-anno").classList.toggle("wrapped");
                }
            });
            '''  # noqa
        doc.extraStyles["style-mdn-anno"] = '''
            @media (max-width: 767px) { .mdn-anno { opacity: .1 } }
            .mdn-anno { font: 1em sans-serif; padding: 0.3em; position: absolute; z-index: 8; right: 0.3em; background: #EEE; color: black; box-shadow: 0 0 3px #999; overflow: hidden; border-collapse: initial; border-spacing: initial; min-width: 9em; max-width: min-content; white-space: nowrap; word-wrap: normal; hyphens: none}
            .mdn-anno:not(.wrapped) { opacity: 1}
            .mdn-anno:hover { z-index: 9 }
            .mdn-anno.wrapped { min-width: 0 }
            .mdn-anno.wrapped > :not(button) { display: none; }
            .mdn-anno > .mdn-anno-btn { cursor: pointer; border: none; color: #000; background: transparent; margin: -8px; float: right; padding: 10px 8px 8px 8px; outline: none; }
            .mdn-anno > .mdn-anno-btn > .less-than-two-engines-flag { color: red; padding-right: 2px; }
            .mdn-anno > .mdn-anno-btn > .all-engines-flag { color: green; padding-right: 2px; }
            .mdn-anno > .mdn-anno-btn > span { color: #fff; background-color: #000; font-weight: normal; font-family: zillaslab, Palatino, "Palatino Linotype", serif; padding: 2px 3px 0px 3px; line-height: 1.3em; vertical-align: top; }
            .mdn-anno > .feature { margin-top: 20px; }
            .mdn-anno > .feature:not(:first-of-type) { border-top: 1px solid #999; margin-top: 6px; padding-top: 2px; }
            .mdn-anno > .feature > .less-than-two-engines-text { color: red }
            .mdn-anno > .feature > .all-engines-text { color: green }
            .mdn-anno > .feature > p { font-size: .75em; margin-top: 6px; margin-bottom: 0; }
            .mdn-anno > .feature > p + p { margin-top: 3px; }
            .mdn-anno > .feature > .support { display: block; font-size: 0.6em; margin: 0; padding: 0; margin-top: 2px }
            .mdn-anno > .feature > .support + div { padding-top: 0.5em; }
            .mdn-anno > .feature > .support > hr { display: block; border: none; border-top: 1px dotted #999; padding: 3px 0px 0px 0px; margin: 2px 3px 0px 3px; }
            .mdn-anno > .feature > .support > hr::before { content: ""; }
            .mdn-anno > .feature > .support > span { padding: 0.2em 0; display: block; display: table; }
            .mdn-anno > .feature > .support > span.no { color: #CCCCCC; filter: grayscale(100%); }
            .mdn-anno > .feature > .support > span.no::before { opacity: 0.5; }
            .mdn-anno > .feature > .support > span:first-of-type { padding-top: 0.5em; }
            .mdn-anno > .feature > .support > span > span { padding: 0 0.5em; display: table-cell; }
            .mdn-anno > .feature > .support > span > span:first-child { width: 100%; }
            .mdn-anno > .feature > .support > span > span:last-child { width: 100%; white-space: pre; padding: 0; }
            .mdn-anno > .feature > .support > span::before { content: ' '; display: table-cell; min-width: 1.5em; height: 1.5em; background: no-repeat center center; background-size: contain; text-align: right; font-size: 0.75em; font-weight: bold; }
            .mdn-anno > .feature > .support > .chrome_android::before { background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg); }
            .mdn-anno > .feature > .support > .firefox_android::before { background-image: url(https://resources.whatwg.org/browser-logos/firefox.png); }
            .mdn-anno > .feature > .support > .chrome::before { background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg); }
            .mdn-anno > .feature > .support > .edge_blink::before { background-image: url(https://resources.whatwg.org/browser-logos/edge.svg); }
            .mdn-anno > .feature > .support > .edge::before { background-image: url(https://resources.whatwg.org/browser-logos/edge_legacy.svg); }
            .mdn-anno > .feature > .support > .firefox::before { background-image: url(https://resources.whatwg.org/browser-logos/firefox.png); }
            .mdn-anno > .feature > .support > .ie::before { background-image: url(https://resources.whatwg.org/browser-logos/ie.png); }
            .mdn-anno > .feature > .support > .safari_ios::before { background-image: url(https://resources.whatwg.org/browser-logos/safari-ios.svg); }
            .mdn-anno > .feature > .support > .nodejs::before { background-image: url(https://nodejs.org/static/images/favicons/favicon.ico); }
            .mdn-anno > .feature > .support > .opera_android::before { background-image: url(https://resources.whatwg.org/browser-logos/opera.svg); }
            .mdn-anno > .feature > .support > .opera::before { background-image: url(https://resources.whatwg.org/browser-logos/opera.svg); }
            .mdn-anno > .feature > .support > .safari::before { background-image: url(https://resources.whatwg.org/browser-logos/safari.png); }
            .mdn-anno > .feature > .support > .samsunginternet_android::before { background-image: url(https://resources.whatwg.org/browser-logos/samsung.svg); }
            .mdn-anno > .feature > .support > .webview_android::before { background-image: url(https://resources.whatwg.org/browser-logos/android-webview.png); }
            .name-slug-mismatch { color: red }
            .caniuse-status:hover { z-index: 9; }

            /* dt, li, .issue, .note, and .example are "position: relative", so to put annotation at right margin, must move to right of containing block */
            .h-entry:not(.status-LS) dt > .mdn-anno, .h-entry:not(.status-LS) li > .mdn-anno, .h-entry:not(.status-LS) .issue > .mdn-anno, .h-entry:not(.status-LS) .note > .mdn-anno, .h-entry:not(.status-LS) .example > .mdn-anno { right: -6.7em; }
            .h-entry p + .mdn-anno { margin-top: 0; }
            h2 + .mdn-anno.after { margin: -48px 0 0 0; }
            h3 + .mdn-anno.after { margin: -46px 0 0 0; }
            h4 + .mdn-anno.after { margin: -42px 0 0 0; }
            h5 + .mdn-anno.after { margin: -40px 0 0 0; }
            h6 + .mdn-anno.after { margin: -40px 0 0 0; }
            '''  # noqa

def createAnno(className, mdnButton, featureDivs):
    return E.aside({"class": className}, mdnButton, featureDivs)

def panelsFromData(doc, data):
    mdnBaseUrl = "https://developer.mozilla.org/en-US/docs/Web/"
    bcdBaseUrl = "https://github.com/mdn/browser-compat-data/blob/master/"

    browsersProvidingCurrentEngines = ["firefox", "safari", "chrome"]
    browsersWithBorrowedEngines = ["opera", "edge_blink"]
    browsersWithRetiredEngines = ["edge", "ie"]
    browsersForMobileDevices = ["firefox_android", "safari_ios",
                                "chrome_android", "webview_android",
                                "samsunginternet_android", "opera_android"]

    # BCD/mdn-spec-links shortnames to full names
    nameFromCodeName = {
        "chrome": "Chrome",
        "chrome_android": "Chrome for Android",
        "edge": "Edge (Legacy)",
        "edge_blink": "Edge",
        "firefox": "Firefox",
        "firefox_android": "Firefox for Android",
        "ie": "IE",
        "nodejs": "Node.js",
        "opera": "Opera",
        "opera_android": "Opera Mobile",
        "safari": "Safari",
        "samsunginternet_android": "Samsung Internet",
        "safari_ios": "iOS Safari",
        "webview_android": "Android WebView"
    }

    panels = False
    for elementId, features in data.items():
        lessThanTwoEngines = 0
        onlyTwoEngines = 0
        allEngines = 0
        featureDivs = []
        targetElement = find(f"[id='{elementId}']", doc)
        if targetElement is None:
            warn(f"No '{elementId}' ID found, skipping MDN features that would target it.")
            continue
        else:
            panels = True
            if targetElement.tag in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
                isAnnoForHeadingContent = True
            else:
                for ancestor in targetElement.iterancestors():
                    if ancestor.tag in ['body', 'main', 'article', 'aside',
                                        'nav', 'section', 'header', 'footer']:
                        break
                    targetElement = ancestor
                    if ancestor.tag in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
                        isAnnoForHeadingContent = True
                        break
                    if ancestor.tag in ['td', 'dt', 'dd', 'li']:
                        isAnnoForListItemOrTableContent = True
                        break
                    if ancestor.tag in ['pre', 'xmp', 'p']:
                        break
        for feature in features:
            isAnnoForHeadingContent = False
            isAnnoForListItemOrTableContent = False
            if "engines" in feature:
                engines = len(feature["engines"])
                if engines < 2:
                    lessThanTwoEngines = lessThanTwoEngines + 1
                elif engines == 2:
                    onlyTwoEngines = onlyTwoEngines + 1
                elif engines >= len(browsersProvidingCurrentEngines):
                    allEngines = allEngines + 1
            featureDivs.append(
                mdnPanelFor(
                    feature,
                    mdnBaseUrl,
                    bcdBaseUrl,
                    nameFromCodeName,
                    browsersProvidingCurrentEngines,
                    browsersWithBorrowedEngines,
                    browsersWithRetiredEngines,
                    browsersForMobileDevices))

        mdnButton = E.button({"class": "mdn-anno-btn"})
        if lessThanTwoEngines > 0:
            appendChild(mdnButton,
                E.b({"class": "less-than-two-engines-flag",
                     "title": "This feature is in less than two current engines."},
                    "\u26A0"))
        elif allEngines > 0 and lessThanTwoEngines == 0 and onlyTwoEngines == 0:
            appendChild(mdnButton,
                E.b({"class": "all-engines-flag",
                     "title": "This feature is in all current engines."},
                    "\u2714"))
        appendChild(mdnButton, E.span("MDN"))

        className = "mdn-anno wrapped"
        if isAnnoForListItemOrTableContent:
            if targetElement.getchildren() \
                    and hasClass(targetElement.getchildren()[0], 'mdn-anno'):
                # If there's already an annotation at the point where we want
                # this, just re-use it (instead of creating another one).
                appendChild(targetElement.getchildren()[0], featureDivs)
            else:
                # For elements we're annotating inside a dt, dd, li, or td, we
                # prepend the annotation to the dt, dd, li, or td — because in
                # cases where we have a long table or list, all the annotations
                # for everything in it otherwise ends up being merged into a
                # single annotation way up at the top of the table or list.
                prependChild(targetElement, createAnno(className, mdnButton, featureDivs))
        elif isAnnoForHeadingContent:
            className = "mdn-anno wrapped after"
            if targetElement.getnext() is not None \
                    and targetElement.getnext().get('class') == className:
                # If there's already an annotation at the point where we want
                # this, just re-use it (instead of creating another one).
                appendChild(targetElement.getnext(), featureDivs)
            else:
                # For elements we're annotating inside an h1-h6 heading, we
                # insert the annotation as the next sibling of the heading.
                insertAfter(targetElement, createAnno(className, mdnButton, featureDivs))
        else:
            if targetElement.getprevious() is not None \
                    and targetElement.getprevious().get('class') == className:
                # If there's already an annotation at the point where we want
                # this, just re-use it (instead of creating another one) —
                # unless it's a class=after annotation (following a heading).
                appendChild(targetElement.getprevious(), featureDivs)
            else:
                # For elements we're annotating that aren't inside a table or
                # list or heading, we insert the annotation as the previous
                # sibling of whatever block-level element holds the element.
                insertBefore(targetElement, createAnno(className, mdnButton, featureDivs))
    return panels

def addSupportRow(browserCodeName, nameFromCodeName, support, supportData):
    if browserCodeName not in support:
        return
    isEdgeLegacy = browserCodeName == "edge"
    isIE = browserCodeName == "ie"
    needsFlag = False
    versionAdded = None
    versionRemoved = None
    minVersion = None
    thisBrowserSupport = support[browserCodeName]
    if isinstance(thisBrowserSupport, dict):
        if "version_added" in thisBrowserSupport:
            versionAdded = thisBrowserSupport["version_added"]
            if "flags" in thisBrowserSupport:
                needsFlag = True
            if "prefix" in thisBrowserSupport or \
                    'alternative_name' in thisBrowserSupport or \
                    'partial_implementation' in thisBrowserSupport:
                versionAdded = False
        if "version_removed" in thisBrowserSupport:
            versionRemoved = thisBrowserSupport["version_removed"]
    if isinstance(thisBrowserSupport, list):
        for versionDetails in thisBrowserSupport:
            if "version_removed" in versionDetails:
                versionRemoved = versionDetails["version_removed"]
                continue
            if "version_added" in versionDetails:
                if versionDetails["version_added"] is False:
                    versionAdded = False
                    continue
                if versionDetails["version_added"] is None:
                    versionAdded = None
                    continue
                if "prefix" in versionDetails or \
                        'alternative_name' in versionDetails or \
                        'partial_implementation' in versionDetails:
                    continue
                if "flags" in thisBrowserSupport:
                    needsFlag = True
                versionAdded = versionDetails["version_added"]
                versionRemoved = None
                break
    statusCode = "n"
    if versionAdded is None:
        minVersion = "?"
    elif versionAdded is False:
        minVersion = "None"
    elif versionAdded is True:
        minVersion = "Yes"
        statusCode = "y"
    else:
        if versionRemoved is None:
            statusCode = "y"
            minVersion = versionAdded + "+"
            if (isEdgeLegacy and versionAdded == "18"):
                minVersion = "18"
            if (isIE and versionAdded == "11"):
                minVersion = "11"
        else:
            statusCode = "n"
            if versionAdded is not None:
                minVersion = versionAdded + u"\u2013" + versionRemoved
            else:
                minVersion = "None"
    browserFullName = nameFromCodeName[browserCodeName]
    appendChild(supportData,
        browserCompatSpan(
            browserCodeName,
            browserFullName,
            statusCode,
            minVersion,
            needsFlag))


def mdnPanelFor(feature, mdnBaseUrl, bcdBaseUrl, nameFromCodeName,
                browsersProvidingCurrentEngines, browsersWithBorrowedEngines,
                browsersWithRetiredEngines, browsersForMobileDevices):
    featureDiv = E.div({"class": "feature"})
    featureName = feature.get("name", None)
    if "slug" in feature:
        slug = feature["slug"]
        displaySlug = slug.split("/", 1)[1]
        title = feature.get("summary", "")
        mdnURL = mdnBaseUrl + slug
        appendChild(featureDiv,
            E.p({},
                E.a({"href": mdnURL, "title": title}, displaySlug)))
    if "engines" in feature:
        engines = len(feature["engines"])
        enginesPara = None
        if engines == 0:
            enginesPara = E.p({"class": "less-than-two-engines-text"},
                              "In no current engines.")
        elif engines == 1:
            enginesPara = E.p({"class": "less-than-two-engines-text"},
                              "In only one current engine.")
        elif engines >= len(browsersProvidingCurrentEngines):
            enginesPara = E.p({"class": "all-engines-text"},
                              "In all current engines.")
        if enginesPara is not None:
            appendChild(featureDiv, enginesPara)
    supportData = E.div({"class": "support"})
    appendChild(featureDiv, supportData)
    support = feature["support"]
    for browserCodeName in browsersProvidingCurrentEngines:
        addSupportRow(browserCodeName, nameFromCodeName, support, supportData)
    appendChild(supportData, E.hr())
    for browserCodeName in browsersWithBorrowedEngines:
        addSupportRow(browserCodeName, nameFromCodeName, support, supportData)
    appendChild(supportData, E.hr())
    for browserCodeName in browsersWithRetiredEngines:
        addSupportRow(browserCodeName, nameFromCodeName, support, supportData)
    appendChild(supportData, E.hr())
    for browserCodeName in browsersForMobileDevices:
        addSupportRow(browserCodeName, nameFromCodeName, support, supportData)
    if "nodejs" in support:
        appendChild(supportData, E.hr())
        addSupportRow("nodejs", nameFromCodeName, support, supportData)
    return featureDiv


def browserCompatSpan(browserCodeName, browserFullName, statusCode, minVersion,
                      needsFlag):
    # browserCodeName: e.g. "chrome"
    # browserFullName: e.g. "Chrome for Android"
    minVersionAttributes = {}
    flagSymbol = ""
    if needsFlag:
        flagSymbol = "\U0001f530 "
        minVersionAttributes["title"] = \
            "Requires setting a user preference or runtime flag."
    statusClass = {"y": "yes", "n": "no"}[statusCode]
    outer = E.span({"class": browserCodeName + " " + statusClass})
    appendChild(outer, E.span({}, browserFullName))
    appendChild(outer, E.span(minVersionAttributes, flagSymbol + minVersion))
    return outer



# -*- coding: utf-8 -*-

def verify_python_version():
	import sys
	import platform

	if sys.version_info < (3,):
		print('''Bikeshed has updated to Python 3, but you are trying to run it with
	Python {}. For instructions on upgrading, please check:
	https://tabatkins.github.io/bikeshed/#installing'''.format(platform.python_version()))
		sys.exit(1)

	if sys.version_info < (3,7):
		print('''Bikeshed now requires Python 3.7; you are on {}.
	For instructions on how to set up a pyenv with 3.7, see:
	https://tabatkins.github.io/bikeshed/#installing'''.format(platform.python_version()))
		sys.exit(1)
verify_python_version()

def verify_requirements():
    import os
    import pkg_resources
    import sys
    requirements_file_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), 'requirements.txt')
    if os.path.exists(requirements_file_path):
        requirements_met = True
        with open(requirements_file_path, 'r') as requirements_file:
            requirements = [line for line in requirements_file.read().split('\n') if (not line.strip().startswith('-'))]
            for requirement in pkg_resources.parse_requirements(requirements):
                try:
                    distribution = pkg_resources.get_distribution(requirement.project_name)
                    if (distribution not in requirement):
                        print(f'Package {requirement.project_name} version {distribution.version} is not supported.')
                        requirements_met = False
                except Exception:
                    print(f'Package {requirement.project_name} is not installed.')
                    requirements_met = False
        if (not requirements_met):
            print(f'Run "pip3 install -r {requirements_file_path}" to complete installation')
            sys.exit(1)
verify_requirements()

from .cli import main
from .Spec import Spec
from . import update
from . import config



# -*- coding: utf-8 -*-

import json
from collections import OrderedDict
from datetime import datetime
from . import config
from .html import *
from .messages import *

def addCanIUsePanels(doc):
    # Constructs "Can I Use panels" which show a compatibility data summary
    # for a term's feature.
    if not doc.md.includeCanIUsePanels:
        return

    if not doc.canIUse:
        doc.canIUse = CanIUseManager(dataFile=doc.dataFile)

    lastUpdated = datetime.utcfromtimestamp(doc.canIUse.updated).date().isoformat()

    # e.g. 'iOS Safari' -> 'ios_saf'
    classFromBrowser = doc.canIUse.agents

    atLeastOnePanel = False
    elements = findAll("[caniuse]", doc)
    if not elements:
        return
    validateCanIUseURLs(doc, elements)
    for dfn in elements:
        featId = dfn.get("caniuse")
        if not featId:
            continue
        del dfn.attrib["caniuse"]

        featId = featId.lower()
        if not doc.canIUse.hasFeature(featId):
            die("Unrecognized Can I Use feature ID: {0}", featId, el=dfn)
        feature = doc.canIUse.getFeature(featId)

        addClass(dfn, "caniuse-paneled")
        panel = canIUsePanelFor(id=featId, data=feature, update=lastUpdated, classFromBrowser=classFromBrowser)
        dfnId = dfn.get("id")
        if not dfnId:
            die("Elements with `caniuse` attribute need to have an ID as well. Got:\n{0}", serializeTag(dfn), el=dfn)
            continue
        panel.set("data-dfn-id", dfnId)
        appendChild(doc.body, panel)
        atLeastOnePanel = True

    if atLeastOnePanel:
        doc.extraScripts['script-caniuse-panel'] = '''
            window.addEventListener("load", function(){
                var panels = [].slice.call(document.querySelectorAll(".caniuse-status"));
                for(var i = 0; i < panels.length; i++) {
                    var panel = panels[i];
                    var dfn = document.querySelector("#" + panel.getAttribute("data-dfn-id"));
                    var rect = dfn.getBoundingClientRect();
                    panel.style.top = (window.scrollY + rect.top) + "px";
                }
            });
            document.body.addEventListener("click", function(e) {
                if(e.target.classList.contains("caniuse-panel-btn")) {
                    e.target.parentNode.classList.toggle("wrapped");
                }
            });'''
        doc.extraStyles['style-caniuse-panel'] = '''
            .caniuse-status { font: 1em sans-serif; width: 9em; padding: 0.3em; position: absolute; z-index: 8; top: auto; right: 0.3em; background: #EEE; color: black; box-shadow: 0 0 3px #999; overflow: hidden; border-collapse: initial; border-spacing: initial; }
            .caniuse-status.wrapped { width: 1em; height: 1em; }
            .caniuse-status.wrapped > :not(input) { display: none; }
            .caniuse-status > input { position: absolute; right: 0; top: 0; width: 1em; height: 1em; border: none; background: transparent; padding: 0; margin: 0; }
            .caniuse-status > p { font-size: 0.6em; margin: 0; padding: 0; }
            .caniuse-status > p + p { padding-top: 0.5em; }
            .caniuse-status > .support { display: block; }
            .caniuse-status > .support > span { padding: 0.2em 0; display: block; display: table; }
            .caniuse-status > .support > span.partial { color: #666666; filter: grayscale(50%); }
            .caniuse-status > .support > span.no { color: #CCCCCC; filter: grayscale(100%); }
            .caniuse-status > .support > span.no::before { opacity: 0.5; }
            .caniuse-status > .support > span:first-of-type { padding-top: 0.5em; }
            .caniuse-status > .support > span > span { padding: 0 0.5em; display: table-cell; vertical-align: top; }
            .caniuse-status > .support > span > span:first-child { width: 100%; }
            .caniuse-status > .support > span > span:last-child { width: 100%; white-space: pre; padding: 0; }
            .caniuse-status > .support > span::before { content: ' '; display: table-cell; min-width: 1.5em; height: 1.5em; background: no-repeat center center; background-size: contain; text-align: right; font-size: 0.75em; font-weight: bold; }
            .caniuse-status > .support > .and_chr::before { background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg); }
            .caniuse-status > .support > .and_ff::before { background-image: url(https://resources.whatwg.org/browser-logos/firefox.png); }
            .caniuse-status > .support > .and_uc::before { background-image: url(https://resources.whatwg.org/browser-logos/uc.png); } /* UC Browser for Android */
            .caniuse-status > .support > .android::before { background-image: url(https://resources.whatwg.org/browser-logos/android.svg); }
            .caniuse-status > .support > .bb::before { background-image: url(https://resources.whatwg.org/browser-logos/bb.jpg); } /* Blackberry Browser */
            .caniuse-status > .support > .chrome::before { background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg); }
            .caniuse-status > .support > .edge::before { background-image: url(https://resources.whatwg.org/browser-logos/edge.svg); }
            .caniuse-status > .support > .firefox::before { background-image: url(https://resources.whatwg.org/browser-logos/firefox.png); }
            .caniuse-status > .support > .ie::before { background-image: url(https://resources.whatwg.org/browser-logos/ie.png); }
            .caniuse-status > .support > .ie_mob::before { background-image: url(https://resources.whatwg.org/browser-logos/ie-mobile.svg); }
            .caniuse-status > .support > .ios_saf::before { background-image: url(https://resources.whatwg.org/browser-logos/safari-ios.svg); }
            .caniuse-status > .support > .op_mini::before { background-image: url(https://resources.whatwg.org/browser-logos/opera-mini.png); }
            .caniuse-status > .support > .op_mob::before { background-image: url(https://resources.whatwg.org/browser-logos/opera.svg); }
            .caniuse-status > .support > .opera::before { background-image: url(https://resources.whatwg.org/browser-logos/opera.svg); }
            .caniuse-status > .support > .safari::before { background-image: url(https://resources.whatwg.org/browser-logos/safari.png); }
            .caniuse-status > .support > .samsung::before { background-image: url(https://resources.whatwg.org/browser-logos/samsung.svg); }
            .caniuse-status > .caniuse { text-align: right; font-style: italic; }
            @media (max-width: 767px) {
                .caniuse-status.wrapped { width: 9em; height: auto; }
                .caniuse-status:not(.wrapped) { width: 1em; height: 1em; }
                .caniuse-status.wrapped > :not(input) { display: block; }
                .caniuse-status:not(.wrapped) > :not(input) { display: none; }
            }'''


def canIUsePanelFor(id, data, update, classFromBrowser):
    panel = E.aside({"class": "caniuse-status", "data-deco": ""},
        E.input({"value": "\u22F0", "type": "button", "class":"caniuse-panel-btn"}))
    mainPara = E.p({"class": "support"}, E.b({}, "Support:"))
    appendChild(panel, mainPara)
    for browser,support in data['support'].items():
        statusCode = support[0]
        if statusCode == "u":
            continue
        minVersion = support[2:]
        appendChild(mainPara,
            browserCompatSpan(classFromBrowser[browser], browser, statusCode, minVersion))
    appendChild(panel,
        E.p({"class": "caniuse"},
            "Source: ",
            E.a({"href": "https://caniuse.com/#feat=" + id}, "caniuse.com"),
            " as of " + update))
    return panel


def browserCompatSpan(browserCodeName, browserFullName, statusCode, minVersion=None):
    if statusCode == "n" or minVersion is None:
        minVersion = "None"
    elif minVersion == "all":
        minVersion = "All"
    else:
        minVersion = minVersion + "+"
    # browserCodeName: e.g. and_chr, ios_saf, ie, etc...
    # browserFullName: e.g. "Chrome for Android"
    statusClass = {"y": "yes", "n": "no", "a": "partial"}[statusCode]
    outer = E.span({"class": browserCodeName + " " + statusClass})
    if statusCode == "a":
        appendChild(outer,
            E.span({},
                E.span({},
                    browserFullName,
                    " (limited)")))
    else:
        appendChild(outer,
            E.span({}, browserFullName))
    appendChild(outer,
        E.span({},
            minVersion))
    return outer


def validateCanIUseURLs(doc, elements):
    # First, ensure that each Can I Use URL shows up at least once in the data;
    # otherwise, it's an error to be corrected somewhere.
    urlFeatures = set()
    for url in doc.md.canIUseURLs:
        sawTheURL = False
        for featureID, featureUrl in doc.canIUse.urlFromFeature.items():
            if featureUrl.startswith(url):
                sawTheURL = True
                urlFeatures.add(featureID)
        if not sawTheURL and url not in doc.md.ignoreCanIUseUrlFailure:
            warn("The Can I Use URL '{0}' isn't associated with any of the Can I Use features. Please check Can I Use for the correct spec url, and either correct your spec or correct Can I Use. If the URL is correct and you'd like to keep it in pre-emptively, add the URL to a 'Ignore Can I Use URL Failure' metadata.", url)

    # Second, ensure that every feature in the data corresponding to one of the listed URLs
    # has a corresponding Can I Use entry in the document;
    # otherwise, you're missing some features.
    docFeatures = set()
    for el in elements:
        featureID = el.get("caniuse").lower()
        docFeatures.add(featureID)

    unusedFeatures = urlFeatures - docFeatures
    if unusedFeatures:
        warn("The following Can I Use features are associated with your URLs, but don't show up in your spec:\n{0}",
             "\n".join(" * {0} - https://caniuse.com/#feat={0}".format(x) for x in sorted(unusedFeatures)))

class CanIUseManager(object):

    def __init__(self, dataFile):
        self.dataFile = dataFile
        data = json.loads(self.dataFile.fetch("caniuse", "data.json", str=True), object_pairs_hook=OrderedDict)
        self.updated = data["updated"]
        self.agents = data["agents"]
        self.urlFromFeature = data["features"]
        self.features = {}

    def hasFeature(self, featureName):
        return featureName in self.urlFromFeature

    def getFeature(self, featureName):
        if featureName in self.features:
            return self.features[featureName]
        if not self.hasFeature(featureName):
            return
        data = json.loads(self.dataFile.fetch("caniuse", "feature-{0}.json".format(featureName), str=True), object_pairs_hook=OrderedDict)
        self.features[featureName] = data
        return data



# -*- coding: utf-8 -*-


from .accidental2119 import accidental2119
from .brokenLinks import brokenLinks
from .exampleIDs import exampleIDs
from .missingExposed import missingExposed
from .requiredIDs import requiredIDs
from .unusedInternalDfns import unusedInternalDfns



# -*- coding: utf-8 -*-

import io
import json
import re
import os
import requests
from collections import defaultdict

from .. import biblio
from ..DefaultOrderedDict import DefaultOrderedDict
from ..messages import *



def update(path, dryRun=False):
    say("Downloading biblio data...")
    biblios = defaultdict(list)
    biblio.processSpecrefBiblioFile(getSpecrefData(), biblios, order=3)
    biblio.processSpecrefBiblioFile(getWG21Data(), biblios, order=3)
    biblio.processReferBiblioFile(getCSSWGData(), biblios, order=4)
    writtenPaths = set()
    if not dryRun:
        groupedBiblios, allNames = groupBiblios(biblios)
        # Save each group to a file
        for group, biblios in groupedBiblios.items():
            try:
                p = os.path.join(path, "biblio", "biblio-{0}.data".format(group))
                writtenPaths.add(p)
                with io.open(p, 'w', encoding="utf-8") as fh:
                    writeBiblioFile(fh, biblios)
            except Exception as e:
                die("Couldn't save biblio database to disk.\n{0}", e)
                return

        # biblio-keys is used to help correct typos,
        # so remove all the useless purely-numbered name types,
        # as they aren't useful for typo-correction
        reducedNames = []
        for name in allNames:
            if re.search(r"-\d{8}$", name):
                continue
            if re.match(r"cwg\d+$", name):
                continue
            if re.match(r"ewg\d+$", name):
                continue
            if re.match(r"fs\d+$", name):
                continue
            if re.match(r"lewg\d+$", name):
                continue
            if re.match(r"lwg\d+$", name):
                continue
            if re.match(r"n\d+$", name):
                continue
            if re.match(r"p\d+(r\d+)?$", name):
                continue
            if re.match(r"d\d+(r\d+)?$", name):
                continue
            if re.match(r"rfc\d+$", name):
                continue
            if re.match(r"wg21-", name):
                continue
            if re.match(r"edit\d+$", name):
                continue
            if re.match(r"etsi-", name):
                continue
            if re.match(r"iso-\d+", name):
                continue
            if re.match(r"iso\d+", name):
                continue
            if re.match(r"scte\d+", name):
                continue
            reducedNames.append(name)
        try:
            p = os.path.join(path, "biblio-keys.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as fh:
                fh.write(str(json.dumps(reducedNames, indent=0, ensure_ascii=False, sort_keys=True)))
        except Exception as e:
            die("Couldn't save biblio database to disk.\n{0}", e)
            return

        # Collect all the number-suffix names which also exist un-numbered
        numberedNames = collectNumberedNames(reducedNames)
        try:
            p = os.path.join(path, "biblio-numeric-suffixes.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as fh:
                fh.write(str(json.dumps(numberedNames, indent=0, ensure_ascii=False, sort_keys=True)))
        except Exception as e:
            die("Couldn't save biblio numeric-suffix information to disk.\n{0}", e)
    say("Success!")
    return writtenPaths


def getSpecrefData():
    try:
        return requests.get("https://api.specref.org/bibrefs").text
    except requests.exceptions.RequestException:
        # SpecRef uses SNI, which old Pythons (pre-2.7.10) don't understand.
        # First try the older herokuapp.com URL.
        try:
            return requests.get("https://specref.herokuapp.com/bibrefs").text
        except:
            # Try the CSSWG proxy.
            try:
                return requests.get("https://api.csswg.org/bibrefs").text
            except:
                warn("Your Python is too old (pre-2.7.10) to talk to SpecRef over HTTPS, and something's wrong with the CSSWG proxy as well. Report this to the Bikeshed repo, please?")
                raise
                return "{}"
    except Exception as e:
        die("Couldn't download the SpecRef biblio data.\n{0}", e)
        return "{}"


def getWG21Data():
    try:
        return requests.get("https://wg21.link/specref.json").text
    except Exception as e:
        die("Couldn't download the WG21 biblio data.\n{0}", e)
        return "{}"


def getCSSWGData():
    try:
        return requests.get("https://raw.githubusercontent.com/w3c/csswg-drafts/master/biblio.ref").text.splitlines()
    except Exception as e:
        die("Couldn't download the CSSWG biblio data.\n{0}", e)
        return []


def groupBiblios(biblios):
    # Group the biblios by the first two letters of their keys
    groupedBiblios = DefaultOrderedDict(DefaultOrderedDict)
    allNames = []
    for k,v in sorted(biblios.items(), key=lambda x:x[0].lower()):
        allNames.append(k)
        group = k[0:2]
        groupedBiblios[group][k] = v
    return groupedBiblios, allNames


def writeBiblioFile(fh, biblios):
    '''
    Each line is a value for a specific key, in the order:

    key
    linkText
    date
    status
    title
    snapshot_url
    current_url
    obsoletedBy
    other
    etAl (as a boolish string)
    authors* (each on a separate line, an indeterminate number of lines)

    Each entry (including last) is ended by a line containing a single - character.
    '''
    typePrefixes = {
        "dict": "d",
        "string": "s",
        "alias": "a"
    }
    for key, entries in biblios.items():
        b = sorted(entries, key=lambda x:x['order'])[0]
        format = b['biblioFormat']
        fh.write("{prefix}:{key}\n".format(prefix=typePrefixes[format], key=key.lower()))
        if format == "dict":
            for field in ["linkText", "date", "status", "title", "snapshot_url", "current_url", "obsoletedBy", "other"]:
                fh.write(b.get(field, "") + "\n")
            if b.get("etAl", False):
                fh.write("1\n")
            else:
                fh.write("\n")
            for author in b.get("authors", []):
                fh.write(author + "\n")
        elif format == "string":
            fh.write(b['linkText'] + "\n")
            fh.write(b['data'] + "\n")
        elif format == "alias":
            fh.write(b['linkText'] + "\n")
            fh.write(b['aliasOf'] + "\n")
        else:
            die("The biblio key '{0}' has an unknown biblio type '{1}'.", key, format)
            continue
        fh.write("-" + "\n")

def collectNumberedNames(names):
    '''
    Collects the set of names that have numeric suffixes
    (excluding ones that look like dates)
    for better error-correction.
    '''

    names = set(names)
    prefixes = defaultdict(list)
    for name in names:
        # Ignoring 4+ digits, as they're probably years or isodates.
        match = re.match(r"(.+\D)\d{1,3}$", name)
        if match:
            prefix = match.group(1)
            if prefix.endswith("-"):
                prefix = prefix[:-1]
            if prefix in names:
                prefixes[prefix].append(name)
                prefixes[prefix] = sorted(prefixes[prefix])
    return prefixes



# -*- coding: utf-8 -*-


from ..html import *
from ..messages import *

def requiredIDs(doc):
	if not doc.md.requiredIDs:
		return
	doc_ids = {e.get("id") for e in findAll("[id]", doc)}
	for id in doc.md.requiredIDs:
		if id.startswith("#"):
			id = id[1:]
		if id not in doc_ids:
			die("Required ID '{0}' was not found in the document.", id)
